from mmdnn.conversion.common.DataStructure.emitter import Emitter
from mmdnn.conversion.common.IR.IR_graph import IRGraph
import os.path
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
import numpy as np
import sys


class OnnxEmitter(Emitter):
    dtype_map = {
        graph_pb2.DT_FLOAT32: "TensorProto.FLOAT"
    }

    transpose_map = {
        1: 2,
        2: 3,
        -1: 1
    }

    def __init__(self, architecture, weight):
        super(OnnxEmitter, self).__init__()
        if os.path.exists(architecture) == False:
            raise ValueError("IR architecture file [{}] is not found.".format(architecture))
        else:
            self.IR_graph = IRGraph(architecture)
            self.IR_graph.build()

        if os.path.exists(weight) == False:
            raise ValueError("IR weight file [{}] is not found.".format(weight))
        else:
            self._load_weights(weight)

    @property
    def header_code(self):
        return """import numpy as np
from onnx import helper, TensorProto
import onnx

__weights_dict = dict()

def load_weights(weight_file):
    if weight_file == None:
        return

    try:
        weights_dict = np.load(weight_file).item()
    except:
        weights_dict = np.load(weight_file, encoding='bytes').item()

    return weights_dict


def KitModel(weight_file = None):
    global __weights_dict
    __weights_dict = load_weights(weight_file)

"""

    def gen_code(self, phase):
        self.phase = phase
        self.add_body(0, self.header_code)

        self.inputs = []
        self.outputs = []
        self.nodes = []
        self.initializer = []

        for layer in self.IR_graph.topological_sort:
            current_node = self.IR_graph.get_node(layer)
            node_type = current_node.type

            if hasattr(self, "emit_" + node_type):
                func = getattr(self, "emit_" + node_type)
                func(current_node)
            else:
                print("OnnxEmitter has not supported operator [%s]." % (node_type))
                self.emit_UNKNOWN(current_node)

        self._process_output_layers()

        self.add_body(1, "graph = helper.make_graph([{}], 'mmdnn', [{}], [{}], [{}])".format(', '.join(self.nodes),
                                                                                             ', '.join(self.inputs),
                                                                                             ', '.join(self.outputs),
                                                                                             ', '.join(
                                                                                                 self.initializer))
                      )
        self.add_body(1, "return helper.make_model(graph, opset_imports=[helper.make_opsetid('', 6)])")
        return self.body_code

    def run(self, dstNetworkPath, dstWeightPath=None, phase='test'):
        super(OnnxEmitter, self).run(dstNetworkPath, dstWeightPath, phase)
        self.save_weights(self.weights_dict, dstWeightPath)

    def check_if_need_transpose(self, IR_node):
        parent = self.IR_graph.get_parent(IR_node.name, [0])
        while parent.type == 'Flatten' or parent.type == 'Dropout':
            parent = self.IR_graph.get_parent(parent.name, [0])
        dim = len(parent.layer.attr['_output_shapes'].list.shape[0].dim)
        if dim > 2:
            original_dims = self.weights_dict[IR_node.name]['weights'].shape
            dims = [i.size for i in parent.layer.attr['_output_shapes'].list.shape[0].dim[1:]] + [-1]
            self.weights_dict[IR_node.name]['weights'] = self.weights_dict[IR_node.name]['weights']
            self.weights_dict[IR_node.name]['weights'] = np.reshape(self.weights_dict[IR_node.name]['weights'], dims)
            self.weights_dict[IR_node.name]['weights'] = np.transpose(self.weights_dict[IR_node.name]['weights'], [dim - 2] + list(range(0, dim - 2)) + [dim - 1])
            self.weights_dict[IR_node.name]['weights'] = np.reshape(self.weights_dict[IR_node.name]['weights'], original_dims)

    def _process_output_layers(self):
        for name in self.IR_graph.output_layers:
            IR_node = self.IR_graph.get_node(self.IR_graph.get_node(name).real_name)
            # omit node of some type
            if IR_node.type == 'Shape' or IR_node.type == 'Pack':
                continue
            shape_str = IRGraph.shapeToStr(IR_node.layer.attr["_output_shapes"].list.shape[0])
            if IR_node.layer.attr['dtype'].type == graph_pb2.DT_UNDEFINED:
                IR_node.layer.attr['dtype'].type = graph_pb2.DT_FLOAT32
            dtype_str = self.dtype_map[IR_node.layer.attr['dtype'].type]
            self.add_body(1, "{:<15} = helper.make_tensor_value_info('{}', {}, ({},))".format(
                IR_node.variable_name + '_out',
                IR_node.variable_name,
                dtype_str,
                shape_str))
            self.outputs.append(IR_node.variable_name + '_out')

    def emit_DataInput(self, IR_node):
        shape = [dim.size if dim.size != -1 else 1 for dim in IR_node.IR_layer.attr["shape"].shape.dim]
        shape_str = ', '.join('%s' % i for i in shape)
        if IR_node.layer.attr['dtype'].type == graph_pb2.DT_UNDEFINED:
            IR_node.layer.attr['dtype'].type = graph_pb2.DT_FLOAT32
        dtype_str = self.dtype_map[IR_node.layer.attr['dtype'].type]
        self.add_body(1, "{:<15} = helper.make_tensor_value_info('{}', {}, ({},))".format(
            IR_node.variable_name + '_orig',
            IR_node.variable_name + '_orig',
            dtype_str,
            shape_str))
        self.add_body(1, "{:15} = helper.make_node('Transpose', inputs=['{}'], outputs=['{}'], perm=[0, 3, 1, 2])".format(
            IR_node.variable_name,
            IR_node.variable_name + '_orig',
            IR_node.variable_name))
        self.inputs.append(IR_node.variable_name + '_orig')
        self.nodes.append(IR_node.variable_name)

    def emit_Conv(self, IR_node):
        kernel_shape = list(IR_node.get_attr('kernel_shape'))[:-2]
        dilations = list(IR_node.get_attr('dilations', [1] * (len(kernel_shape) + 2)))[1:-1]
        group = IR_node.get_attr('group', 1)
        if IR_node.type == 'DepthwiseConv':
            group = IR_node.IR_layer.attr["kernel_shape"].list.i[-2]
            self.weights_dict[IR_node.name]['weights'] = np.swapaxes(self.weights_dict[IR_node.name]['weights'], -1, -2)
        pads = IR_node.get_attr('pads')
        pad_length = len(pads)
        pads = pads[1:pad_length // 2 - 1] + pads[pad_length // 2 + 1:pad_length - 1]
        strides = list(IR_node.get_attr('strides'))[1:-1]
        use_bias=IR_node.get_attr('use_bias')
        self.add_body(1, "{:15} = __weights_dict['{}']['weights']".format(
            IR_node.variable_name + '_weight_array',
            IR_node.name))
        self.add_body(1, "{} = {}.transpose([3,2,0,1])".format(
            IR_node.variable_name + '_weight_array',
            IR_node.variable_name + '_weight_array'))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}.flatten().astype(float)))".format(
                          IR_node.variable_name + '_weight',
                          IR_node.variable_name + '_weight',
                          IR_node.variable_name + '_weight_array',
                          IR_node.variable_name + '_weight_array',
                          IR_node.variable_name + '_weight_array'))
        if use_bias:
            self.add_body(1, "{:15} = __weights_dict['{}']['bias']".format(
                IR_node.variable_name + '_bias_array',
                IR_node.name))
            self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}.flatten().astype(float)))".format(
                              IR_node.variable_name + '_bias',
                              IR_node.variable_name + '_bias',
                              IR_node.variable_name + '_bias_array',
                              IR_node.variable_name + '_bias_array',
                              IR_node.variable_name + '_bias_array'))
            self.add_body(1, "{:15} = helper.make_node('Conv', inputs=['{}', '{}', '{}'],outputs=['{}'], dilations={}, group={}, kernel_shape={}, pads={}, strides={})".format(
                              IR_node.variable_name,
                              self.parent_variable_name(IR_node),
                              IR_node.variable_name + '_weight',
                              IR_node.variable_name + '_bias',
                              IR_node.variable_name,
                              dilations,
                              group,
                              kernel_shape,
                              pads,
                              strides))
            self.nodes.append(IR_node.variable_name + '_bias')
        else:
            self.add_body(1, "{:15} = helper.make_node('Conv', inputs=['{}', '{}'],outputs=['{}'], dilations={}, group={}, kernel_shape={}, pads={}, strides={})".format(
                              IR_node.variable_name,
                              self.parent_variable_name(IR_node),
                              IR_node.variable_name + '_weight',
                              IR_node.variable_name,
                              dilations,
                              group,
                              kernel_shape,
                              pads,
                              strides))
        self.nodes.append(IR_node.variable_name + '_weight')
        self.nodes.append(IR_node.variable_name)

    def emit_BatchNorm(self, IR_node):
        epsilon = IR_node.get_attr('epsilon')
        if IR_node.get_attr('scale'):
            self.add_body(1, "{:15} = __weights_dict['{}']['scale']".format(
                IR_node.variable_name + '_scale_array',
                IR_node.name))
        else:
            self.add_body(1, "{:15} = np.ndarray(__weights_dict['{}']['bias'].shape, dtype=__weights_dict['{}']['bias'].dtype)".format(
                              IR_node.variable_name + '_scale_array',
                              IR_node.name,
                              IR_node.name))
            self.add_body(1, "{:15}.fill(1)".format(IR_node.variable_name + '_scale_array'))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                          IR_node.variable_name + '_scale',
                          IR_node.variable_name + '_scale',
                          IR_node.variable_name + '_scale_array',
                          IR_node.variable_name + '_scale_array',
                          IR_node.variable_name + '_scale_array'))
        self.add_body(1, "{:15} = __weights_dict['{}']['bias']".format(
            IR_node.variable_name + '_bias_array',
            IR_node.name))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                          IR_node.variable_name + '_bias',
                          IR_node.variable_name + '_bias',
                          IR_node.variable_name + '_bias_array',
                          IR_node.variable_name + '_bias_array',
                          IR_node.variable_name + '_bias_array'))
        self.add_body(1, "{:15} = __weights_dict['{}']['mean']".format(
            IR_node.variable_name + '_mean_array',
            IR_node.name))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                          IR_node.variable_name + '_mean',
                          IR_node.variable_name + '_mean',
                          IR_node.variable_name + '_mean_array',
                          IR_node.variable_name + '_mean_array',
                          IR_node.variable_name + '_mean_array'))
        self.add_body(1, "{:15} = __weights_dict['{}']['var']".format(
                          IR_node.variable_name + '_var_array',
                          IR_node.name))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                          IR_node.variable_name + '_var',
                          IR_node.variable_name + '_var',
                          IR_node.variable_name + '_var_array',
                          IR_node.variable_name + '_var_array',
                          IR_node.variable_name + '_var_array'))
        self.add_body(1, "{:15} = helper.make_node('BatchNormalization', inputs=['{}', '{}', '{}', '{}', '{}'],outputs=['{}'], epsilon={}, is_test={})".format(
                          IR_node.variable_name,
                          self.parent_variable_name(IR_node),
                          IR_node.variable_name + '_scale',
                          IR_node.variable_name + '_bias',
                          IR_node.variable_name + '_mean',
                          IR_node.variable_name + '_var',
                          IR_node.variable_name,
                          epsilon,
                          0 if self.phase == 'train' else 1))
        self.nodes.append(IR_node.variable_name + '_scale')
        self.nodes.append(IR_node.variable_name + '_bias')
        self.nodes.append(IR_node.variable_name + '_mean')
        self.nodes.append(IR_node.variable_name + '_var')
        self.nodes.append(IR_node.variable_name)


    def emit_Scale(self, IR_node):
        dims = [i.size for i in IR_node.layer.attr['_output_shapes'].list.shape[0].dim[1:]]
        units = dims[-1]
        epsilon = 1e-5
        if IR_node.get_attr('scale'):
            self.add_body(1, "{:15} = __weights_dict['{}']['scale']".format(
                IR_node.variable_name + '_scale_array',
                IR_node.name))
        else:
            self.add_body(1, "{:15} = np.ndarray(__weights_dict['{}']['bias'].shape, dtype=__weights_dict['{}']['bias'].dtype)".format(
                              IR_node.variable_name + '_scale_array',
                              IR_node.name,
                              IR_node.name))
            self.add_body(1, "{:15}.fill(1)".format(IR_node.variable_name + '_scale_array'))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                          IR_node.variable_name + '_scale',
                          IR_node.variable_name + '_scale',
                          IR_node.variable_name + '_scale_array',
                          IR_node.variable_name + '_scale_array',
                          IR_node.variable_name + '_scale_array'))
        self.add_body(1, "{:15} = __weights_dict['{}']['bias']".format(
            IR_node.variable_name + '_bias_array',
            IR_node.name))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                          IR_node.variable_name + '_bias',
                          IR_node.variable_name + '_bias',
                          IR_node.variable_name + '_bias_array',
                          IR_node.variable_name + '_bias_array',
                          IR_node.variable_name + '_bias_array'))
        self.add_body(1, "{:15} = np.zeros({}, dtype=np.float32)".format(
            IR_node.variable_name + '_mean_array',
            units))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                          IR_node.variable_name + '_mean',
                          IR_node.variable_name + '_mean',
                          IR_node.variable_name + '_mean_array',
                          IR_node.variable_name + '_mean_array',
                          IR_node.variable_name + '_mean_array'))
        self.add_body(1, "{:15} = np.ones({}, dtype=np.float32)".format(
                          IR_node.variable_name + '_var_array',
                          units))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                          IR_node.variable_name + '_var',
                          IR_node.variable_name + '_var',
                          IR_node.variable_name + '_var_array',
                          IR_node.variable_name + '_var_array',
                          IR_node.variable_name + '_var_array'))
        self.add_body(1, "{:15} = helper.make_node('BatchNormalization', inputs=['{}', '{}', '{}', '{}', '{}'],outputs=['{}'], epsilon={}, is_test={})".format(
                          IR_node.variable_name,
                          self.parent_variable_name(IR_node),
                          IR_node.variable_name + '_scale',
                          IR_node.variable_name + '_bias',
                          IR_node.variable_name + '_mean',
                          IR_node.variable_name + '_var',
                          IR_node.variable_name,
                          epsilon,
                          0 if self.phase == 'train' else 1))
        self.nodes.append(IR_node.variable_name + '_scale')
        self.nodes.append(IR_node.variable_name + '_bias')
        self.nodes.append(IR_node.variable_name + '_mean')
        self.nodes.append(IR_node.variable_name + '_var')
        self.nodes.append(IR_node.variable_name)


    def emit_Relu(self, IR_node):
        self.add_body(1, "{:15} = helper.make_node('Relu', inputs=['{}'], outputs=['{}'])".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.variable_name))
        self.nodes.append(IR_node.variable_name)

    def emit_Add(self, IR_node):
        input_layers = ', '.join(
            ("'" + self.IR_graph.get_parent(IR_node.name, [num]).real_variable_name) + "'" for num in
            range(0, len(IR_node.in_edges)))
        self.add_body(1, "{:15} = helper.make_node('Add', inputs=[{}], outputs=['{}'])".format(
            IR_node.variable_name,
            input_layers,
            IR_node.variable_name))
        self.nodes.append(IR_node.variable_name)

    def emit_Pool(self, IR_node):
        pooling_type = IR_node.get_attr('pooling_type')
        if IR_node.layer.attr['global_pooling'].b:
            if pooling_type == 'AVG':
                self.add_body(1, "{:15} = helper.make_node('GlobalAveragePool', inputs=['{}'], outputs=['{}'])".format(
                    IR_node.variable_name,
                    self.parent_variable_name(IR_node),
                    IR_node.variable_name))
                self.nodes.append(IR_node.variable_name)
            else:
                print("OnnxEmitter has not supported Global Pool type [%s]." % (pooling_type))
                self.emit_UNKNOWN(IR_node)
        else:
            if pooling_type in ['AVG', 'MAX']:
                if pooling_type == 'AVG':
                    op_name = 'AveragePool'
                elif pooling_type == 'MAX':
                    op_name = 'MaxPool'
                kernel_shape = list(IR_node.get_attr('kernel_shape')[1:-1])
                pads = IR_node.get_attr('pads')
                pad_length = len(pads)
                pads = pads[1:pad_length // 2 - 1] + pads[pad_length // 2 + 1:pad_length - 1]
                strides = list(IR_node.get_attr('strides')[1:-1])
                self.add_body(1, "{:15} = helper.make_node('{}', inputs=['{}'],outputs=['{}'], kernel_shape={}, pads={}, strides={})".format(
                                  IR_node.variable_name,
                                  op_name,
                                  self.parent_variable_name(IR_node),
                                  IR_node.variable_name,
                                  kernel_shape,
                                  pads,
                                  strides))
                self.nodes.append(IR_node.variable_name)
            else:
                print("OnnxEmitter has not supported Pool type [%s]." % (pooling_type))
                self.emit_UNKNOWN(IR_node)

    def emit_FullyConnected(self, IR_node):
        self.check_if_need_transpose(IR_node)
        use_bias = IR_node.get_attr('use_bias', True)
        units = IR_node.get_attr('units')

        self.add_body(1, "{:15} = __weights_dict['{}']['weights']".format(
            IR_node.variable_name + '_weight_array',
            IR_node.name))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}.flatten().astype(float)))".format(
                          IR_node.variable_name + '_weight',
                          IR_node.variable_name + '_weight',
                          IR_node.variable_name + '_weight_array',
                          IR_node.variable_name + '_weight_array',
                          IR_node.variable_name + '_weight_array'))
        if use_bias:
            self.add_body(1, "{:15} = __weights_dict['{}']['bias']".format(
                IR_node.variable_name + '_bias_array',
                IR_node.name))
        else:
            self.add_body(1, "{:15} = np.zeros({})".format(
                IR_node.variable_name + '_bias_array',
                units))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}.flatten().astype(float)))".format(
                          IR_node.variable_name + '_bias',
                          IR_node.variable_name + '_bias',
                          IR_node.variable_name + '_bias_array',
                          IR_node.variable_name + '_bias_array',
                          IR_node.variable_name + '_bias_array'))
        self.add_body(1, "{:15} = helper.make_node('Gemm', inputs=['{}', '{}', '{}'],outputs=['{}'])".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.variable_name + '_weight',
            IR_node.variable_name + '_bias',
            IR_node.variable_name))
        self.nodes.append(IR_node.variable_name + '_weight')
        self.nodes.append(IR_node.variable_name + '_bias')
        self.nodes.append(IR_node.variable_name)

    def emit_Pad(self, IR_node):
        mode = IR_node.layer.attr['mode'].s.decode()
        pads = IR_node.get_attr('pads')
        pad_length = len(pads)
        pads = [0, 0] + pads[1:pad_length // 2 - 1] + [0, 0] + pads[pad_length // 2 + 1:pad_length - 1]
        self.add_body(1, "{:15} = helper.make_node('Pad', inputs=['{}'], outputs=['{}'], mode='{}', pads={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.variable_name,
            mode,
            pads))
        self.nodes.append(IR_node.variable_name)

    def emit_Concat(self, IR_node):
        axis = IR_node.get_attr('axis') - 2
        inputs = ', '.join("'" + self.IR_graph.get_node(i).real_variable_name + "'" for i in IR_node.in_edges)
        self.add_body(1, "{:15} = helper.make_node('Concat', inputs=[{}], outputs=['{}'], axis={})".format(
            IR_node.variable_name,
            inputs,
            IR_node.variable_name,
            axis))
        self.nodes.append(IR_node.variable_name)

    def emit_Flatten(self, IR_node):
        self.add_body(1, "{:15} = helper.make_node('Flatten', inputs=['{}'], outputs=['{}'])".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.variable_name))
        self.nodes.append(IR_node.variable_name)

    def emit_Softmax(self, IR_node):
        self.add_body(1, "{:15} = helper.make_node('Softmax', inputs=['{}'], outputs=['{}'])".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.variable_name))
        self.nodes.append(IR_node.variable_name)

    def emit_Constant(self, IR_node):
        if IR_node.get_attr('value'):
            value = 'np.array({}, dtype=np.float32)'.format(IR_node.get_attr('value'))
            self.add_body(1, "{:15} = {}".format(
                IR_node.variable_name + '_value_array',
                value))
        else:
            self.add_body(1, "{:15} = __weights_dict['{}']['value']".format(
                IR_node.variable_name + '_value_array',
                IR_node.name))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}.flatten().astype(float)))".format(
                          IR_node.variable_name,
                          IR_node.variable_name,
                          IR_node.variable_name + '_value_array',
                          IR_node.variable_name + '_value_array',
                          IR_node.variable_name + '_value_array'))
        self.nodes.append(IR_node.variable_name)

    def emit_Sub(self, IR_node):
        inputs = ', '.join("'" + self.IR_graph.get_node(i).real_variable_name + "'" for i in IR_node.in_edges)
        self.add_body(1, "{:15} = helper.make_node('Sub', inputs=[{}], outputs=['{}'], broadcast=1)".format(
            IR_node.variable_name,
            inputs,
            IR_node.variable_name))
        self.nodes.append(IR_node.variable_name)

    def emit_Mul(self, IR_node):
        inputs = ', '.join("'" + self.IR_graph.get_node(i).real_variable_name + "'" for i in IR_node.in_edges)
        
        if IR_node.name in self.weights_dict and 'weights' in self.weights_dict[IR_node.name]:
            self.add_body(1,"{:15} = np.array([__weights_dict['{}']['weights']])".format(
                IR_node.variable_name+'_weight_array',
                IR_node.name
            ))
            self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                    IR_node.variable_name + '_weight',
                    IR_node.variable_name + '_weight',
                    IR_node.variable_name + '_weight_array',
                    IR_node.variable_name + '_weight_array',
                    IR_node.variable_name + '_weight_array'))
            inputs += ', '+''.join("'"+IR_node.variable_name +"_weight'")
            self.nodes.append(IR_node.variable_name+'_weight')
        
        self.add_body(1, "{:15} = helper.make_node('Mul', inputs=[{}], outputs=['{}'], broadcast=1)".format(
            IR_node.variable_name,
            inputs,
            IR_node.variable_name))
        self.nodes.append(IR_node.variable_name)

    def emit_Dropout(self, IR_node):
        self.add_body(1, "{:15} = helper.make_node('Dropout', inputs=['{}'], outputs=['{}'], is_test={}, ratio={})".format(
                          IR_node.variable_name,
                          self.parent_variable_name(IR_node),
                          IR_node.variable_name,
                          0 if self.phase == 'train' else 1,
                          1 - IR_node.get_attr('keep_prob')))
        self.nodes.append(IR_node.variable_name)

    def emit_Squeeze(self, IR_node):
        IR_node.real_name = self.IR_graph.get_node(IR_node.in_edges[0]).real_name

    def emit_ReduceMean(self, IR_node):
        axes = IR_node.layer.attr['axes'].list.i[:]
        axes = ','.join('%s' % OnnxEmitter.transpose_map[i] for i in axes)
        self.add_body(1, "{:15} = helper.make_node('ReduceMean', inputs=['{}'], outputs=['{}'], axes=[{}], keepdims={})".format(
                          IR_node.variable_name,
                          self.parent_variable_name(IR_node),
                          IR_node.variable_name,
                          axes,
                          1 if IR_node.layer.attr['keepdims'].b else 0))
        self.nodes.append(IR_node.variable_name)

    def emit_Reshape(self, IR_node):
        shape = [item if item != -1 else 1 for item in IR_node.get_attr('shape')]
        if len(shape) == 4:
            shape = [shape[i] for i in [0, 3, 1, 2]]
        shape_str = ', '.join('%s' % i for i in shape)
        self.add_body(1, "{:15} = np.array([{}], dtype=np.int64)".format(
            IR_node.variable_name + '_shape_array',
            shape_str
        ))
        self.add_body(1, "{:15} = helper.make_node('Constant', inputs=[], outputs=['{}'], value=helper.make_tensor(name='const_tensor', data_type=onnx.mapping.NP_TYPE_TO_TENSOR_TYPE[{}.dtype], dims={}.shape, vals={}))".format(
                          IR_node.variable_name + '_shape',
                          IR_node.variable_name + '_shape',
                          IR_node.variable_name + '_shape_array',
                          IR_node.variable_name + '_shape_array',
                          IR_node.variable_name + '_shape_array'))
        self.add_body(1, "{:15} = helper.make_node('Reshape', inputs=['{}', '{}'], outputs=['{}'])".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.variable_name + '_shape',
            IR_node.variable_name))
        self.nodes.append(IR_node.variable_name + '_shape')
        self.nodes.append(IR_node.variable_name)

    def emit_LRN(self, IR_node):
        alpha = IR_node.get_attr('alpha')
        beta = IR_node.get_attr('beta')
        bias = IR_node.get_attr('bias', 1.0)
        size = IR_node.get_attr('size') * 2 - 1
        self.add_body(1, "{:15} = helper.make_node('LRN', inputs=['{}'], outputs=['{}'], alpha={}, beta={}, bias={}, size={})".format(
                          IR_node.variable_name,
                          self.parent_variable_name(IR_node),
                          IR_node.variable_name,
                          alpha,
                          beta,
                          bias,
                          size))
        self.nodes.append(IR_node.variable_name)

    def emit_Relu6(self, IR_node):
        self.add_body(1, "{:15} = helper.make_node('Clip', inputs=['{}'], outputs=['{}'], min=0.0, max=6.0)".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.variable_name))
        self.nodes.append(IR_node.variable_name)

    def emit_DepthwiseConv(self, IR_node):
        self.emit_Conv(IR_node)

    def emit_Slice(self, IR_node):
        if self.IR_graph.get_parent(IR_node.name, [0]).type == 'Shape':
            pass
        else:
            starts = IR_node.get_attr('starts')
            starts = [starts[0], starts[-1]] + starts[1:-1]
            ends = IR_node.get_attr('ends')
            ends = [ends[0], ends[-1]] + ends[1:-1]
            ends = [i if i != 0 else sys.maxsize for i in ends]
            self.add_body(1, "{:15} = helper.make_node('Slice', inputs=['{}'], outputs=['{}'], starts={}, ends={})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                IR_node.variable_name,
                starts,
                ends))
            self.nodes.append(IR_node.variable_name)

    def emit_LeakyRelu(self, IR_node):
        alpha = IR_node.get_attr('alpha')
        self.add_body(1, "{:15} = helper.make_node('LeakyRelu', inputs=['{}'], outputs=['{}'], alpha={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.variable_name,
            alpha))
        self.nodes.append(IR_node.variable_name)

    def emit_PRelu(self, IR_node):
      slope = IR_node.get_attr('gamma')
      self.add_body(1, "{:15} = helper.make_node('PRelu', inputs=['{}'], outputs=['{}'], slope={})".format(
          IR_node.variable_name,
          self.parent_variable_name(IR_node),
          IR_node.variable_name,
          slope))
      self.nodes.append(IR_node.variable_name)

    def emit_SpaceToDepth(self, IR_node):
        blocksize = IR_node.get_attr('blocksize')
        self.add_body(1, "{:15} = helper.make_node('SpaceToDepth', inputs=['{}'], outputs=['{}'], blocksize={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.variable_name,
            blocksize))
        self.nodes.append(IR_node.variable_name)

    def emit_UNKNOWN(self, IR_node):
        print(IR_node.IR_layer.name)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import  collections

class GraphNode(object):

    def __init__(self, layer):
        self.in_edges = list()
        self.out_edges = list()
        self.layer = layer
        self.covered = False
        self.real_name = self.name

    @property
    def name(self):
        assert False

    @property
    def variable_name(self):
        return self.real_name.replace('/', '_').replace('-', '_').replace('[','_').replace(']','_')

    @property
    def real_variable_name(self):
        return self.real_name.replace('/', '_').replace('-', '_').replace('[','_').replace(']','_')



class Graph(object):

    def __init__(self, model):
        # key: layer_name    value: keras layer
        self.layer_map = collections.OrderedDict()
        self.input_layers = list()
        self.output_layers = list()
        self.layer_name_map = collections.OrderedDict()
        self.topological_sort = list()
        self.model = model


    def build(self):
        self._make_input_layers()
        self._make_output_layers()
        self._get_topological_sort()


    def rebuild(self):
        self._make_input_layers(True)
        self._make_output_layers()
        self._get_topological_sort()

    def _make_input_layers(self, rebuild=False):
        for name, layer in self.layer_map.items():
            layer.left_in_edges = len(layer.in_edges)
            if len(layer.in_edges) == 0:
                if rebuild:
                    if not layer.get_attr('scope'):
                        self.input_layers.append(name)
                else:
                    self.input_layers.append(name)


    def _make_output_layers(self):
        for name, layer in self.layer_map.items():
            if len(layer.out_edges) == 0:
                self.output_layers.append(name)


    '''get node by its name or tensor name'''
    def get_node(self, name):
        if not name.split(':')[0] in self.layer_map:
            raise IOError("Graph doesn't have node [%s]." % name.split(':')[0])
            return None
        else:
            return self.layer_map[name.split(':')[0]]


    def get_nodes(self):
        return self.layer_map.values()


    def get_son(self, name, path, set_flag = False):
        if name == None: return None
        current_node = self.get_node(name)
        for idx in path:
            if len(current_node.out_edges) <= idx: return None
            son_name = current_node.out_edges[idx].split(':')[0]
            current_node = self.get_node(son_name)
            if set_flag:
                current_node.covered = True
        return current_node


    def get_parent(self, name, path, set_flag = False):
        if name == None: return None
        current_node = self.get_node(name)
        for idx in path:
            if len(current_node.in_edges) <= idx: return None
            parent_name = current_node.in_edges[idx].split(':')[0]
            current_node = self.get_node(parent_name)
            if set_flag:
                current_node.covered = True
        return current_node

    def get_real_parent_name(self, name, path, set_flag = False):
        if name == None: return None
        current_node = self.get_node(name)
        for idx in path:
            if len(current_node.in_edges) <= idx: return None
            parent_name = current_node.in_edges[idx].split(':')[0]
            current_node = self.get_node(parent_name)
            if set_flag:
                current_node.covered = True
        return self.layer_name_map[current_node.name]


    def get_parent_variable_name(self, name, path, set_flag = False):
        if name == None: return None
        current_node = self.get_node(name)
        for idx in path:
            if len(current_node.in_edges) <= idx: return None
            parent_name = current_node.in_edges[idx].split(':')[0]
            current_subscriptor = '' if len(current_node.in_edges[idx].split(':'))==1 else '[{}]'.format(current_node.in_edges[idx].split(':')[1])
            current_node = self.get_node(parent_name)
            if set_flag:
                current_node.covered = True

        return current_node.real_variable_name + current_subscriptor


    # private functions
    def _get_topological_sort(self):
        self.topological_sort = self.input_layers[:]
        idx = 0
        while idx < len(self.topological_sort):
            current_node = self.get_node(self.topological_sort[idx])
            for next_node in current_node.out_edges:
                next_node_info = self.get_node(next_node)
                next_node_info.left_in_edges -= self._check_left_in_edges_num(current_node.name, next_node_info) # one node may connect another node by more than one edge. 
                # next_node_info.left_in_edges -= 1
                if next_node_info.left_in_edges == 0:
                    self.topological_sort.append(next_node)
            idx += 1


    def _make_connection(self, src, dst):
        if (src == dst) or (src not in self.layer_map) or (dst not in self.layer_map):
            if src.split(':')[0] not in self.layer_map:
                print ("Warning: Graph Construct a self-loop node {}. Ignored.".format(src))
                return

        # print ('{} --> {}'.format(src, dst))
        if not dst in self.layer_map[src.split(':')[0]].out_edges:
            self.layer_map[src.split(':')[0]].out_edges.append(dst)
        if not src in self.layer_map[dst].in_edges:
            self.layer_map[dst.split(':')[0]].in_edges.append(src)


    def _check_left_in_edges_num(self, in_node_name, node):
        count = 0
        for in_edge in node.in_edges:
            if in_node_name == in_edge.split(':')[0]:
                count += 1
        return count


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import os

from mmdnn.conversion.common.IR.IR_graph import IRGraph, IRGraphNode
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.DataStructure.emitter import Emitter
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.rewriter.folder import Folder


class TensorflowEmitter(Emitter):

    dtype_map = {
        graph_pb2.DT_FLOAT16 : "tf.float16",
        graph_pb2.DT_FLOAT32 : "tf.float32",
        graph_pb2.DT_FLOAT64 : "tf.float64",
        graph_pb2.DT_INT16 : "tf.int16",
        graph_pb2.DT_INT32 : "tf.int32",
        graph_pb2.DT_INT64 : "tf.int64",
        graph_pb2.DT_UINT8 : "tf.uint8",
        graph_pb2.DT_UINT16 : "tf.uint16"
    }


    @property
    def header_code(self):
        return """import tensorflow as tf

__weights_dict = dict()

is_train = {}

def load_weights(weight_file):
    import numpy as np

    if weight_file == None:
        return

    try:
        weights_dict = np.load(weight_file).item()
    except:
        weights_dict = np.load(weight_file, encoding='bytes').item()

    return weights_dict


def KitModel(weight_file = None):
    global __weights_dict
    __weights_dict = load_weights(weight_file)
""".format(self.trainable)


    def __init__(self, model):
        super(TensorflowEmitter, self).__init__()

        from six import string_types as _string_types
        if isinstance(model, _string_types):
            network_path = model
        else:
            network_path = model[0]
            self._load_weights(model[1])

        self.IR_graph = IRGraph(network_path)
        super(TensorflowEmitter, self)._build()
        
        folder = Folder(self.IR_graph, self.weights_dict)
        folder.fold()

    def gen_code(self, phase):
        self.trainable = (phase == 'train')
        self.add_body(0, self.header_code)

        for layer in self.IR_graph.topological_sort:
            current_node = self.IR_graph.get_node(layer)
            node_type = current_node.type

            if hasattr(self, "emit_" + node_type):
                func = getattr(self, "emit_" + node_type)
                line = func(current_node)
                if line != None:
                    self.add_body(1, line)
            else:
                print("TensorflowEmitter has not supported operator [%s]." % (node_type))
                self.emit_UNKNOWN(current_node)


        self.add_body(1, "return {}, {}".format(
            ', '.join([self.IR_graph.get_node(name).real_variable_name for name in self.IR_graph.input_layers if self.IR_graph.get_node(name).type != 'Const' and not self.IR_graph.get_node(name).get_attr('feed_weights')]),
            ', '.join([self.IR_graph.get_node(name).real_variable_name for name in self.IR_graph.output_layers if self.IR_graph.get_node(name).type != 'Pack' and  self.IR_graph.get_node(name).type !='Shape'])))



        self.add_body(0, "")
        for i in self.used_layers:
            func = getattr(self, "_layer_" + i)
            func()

        self.add_body(0, "")
        for code in self.layers_codes.values():
            self.add_body(0, code)

        return self.body_code


    def parent_variable_name(self, IR_node, path=[0]):
        if not IR_node.in_edges and IR_node.name in self.weights_dict.keys():
            return "tf.constant(__weights_dict['{}']['weights'], name='{}')".format(
                IR_node.name,
                IR_node.name)
        return super(TensorflowEmitter, self).parent_variable_name(IR_node, path)


    @staticmethod
    def _shapeToStr(shapes):
        ret = [dim.size if dim.size != -1 else 'None' for dim in shapes.dim]
        return ', '.join('%s' % i for i in ret)


    def emit_Conv(self, IR_node):
        self.used_layers.add(IR_node.type)
        strides_str = ', '.join('%s' % i for i in IR_node.get_attr('strides')[1:-1])
        input_node, padding = self._defuse_padding(IR_node)
        data_format = IR_node.get_attr('data_format')
        code = "{:<15} = convolution({}, group={}, strides=[{}], padding='{}', name='{}')".format(
            IR_node.variable_name,
            input_node,
            IR_node.get_attr('group', 1),
            strides_str,
            padding,
            IR_node.name)
        return code

    def _defuse_padding(self, IR_node, extra_str=""):
        auto_pad = IR_node.get_attr('auto_pad')
        if auto_pad:
            input_node = self.parent_variable_name(IR_node)
            if auto_pad == 'VALID':
                padding = 'VALID'
            elif auto_pad.startswith("SAME"):
                padding = 'SAME'
            else:
                raise ValueError("Unknown padding type [{}].".format(auto_pad))

            return input_node, padding

        else:
            padding = IR_node.get_attr("pads")
            padding = convert_onnx_pad_to_tf(padding)
            if not is_valid_padding(padding):
                input_node = IR_node.variable_name + '_pad'
                self.add_body(1, "{:<15} = tf.pad({}, paddings = {}{})".format(
                    input_node,
                    self.parent_variable_name(IR_node),
                    padding,
                    extra_str
                    ))
            else:
                input_node = self.parent_variable_name(IR_node)

            return input_node, 'VALID'


    def emit_Constant(self, IR_node):
        if 'dtype' in IR_node.layer.attr:
            dtype_str = "{}".format(self.dtype_map[IR_node.layer.attr['dtype'].type])
        else:
            dtype_str = "tf.float32"
        code = "{:<15} = tf.constant({}, dtype={}, name='{}')".format(
            IR_node.variable_name,
            "__weights_dict['{}']['value']".format(IR_node.name) if IR_node.get_attr('value')== None else IR_node.get_attr('value'),
            dtype_str,
            IR_node.name)

        return code


    def emit_Pool(self, IR_node):
        pooling_type = IR_node.get_attr('pooling_type')
        if pooling_type == 'MAX':
            op = 'max_pool'
            padding_const = ", constant_values=float('-Inf')"
        elif pooling_type == 'AVG':
            op = 'avg_pool'
            padding_const = ""
        else:
            raise ValueError("unknown pooling type [{}].".format(pooling_type))

        arrlen = len(IR_node.get_attr('strides'))
        dim_str = '3d' if arrlen == 5 else ""

        if IR_node.layer.attr['global_pooling'].b:
            code = "{:<15} = tf.nn.{}{}({}, [1] + {}.get_shape().as_list()[1:-1] + [1], strides = [1] * {}, padding = 'VALID', name = '{}')".format(
                IR_node.variable_name,
                op,
                dim_str,
                self.parent_variable_name(IR_node),
                self.parent_variable_name(IR_node),
                arrlen,
                IR_node.name)
        else:
            dim = len(IR_node.get_attr("strides")) - 2
            dilations = IR_node.get_attr('dilations')
            if dilations:
                for e in IR_node.get_attr('dilations'):
                    assert e == 1

            pool_size = IR_node.get_attr('kernel_shape')[1:-1]
            strides = IR_node.get_attr('strides')[1:-1]
            padding = IR_node.get_attr('pads')[1:dim]

            if pooling_type == "AVG" and pool_size.count(pool_size[0]) == len(pool_size) and strides[0] == 1 and strides.count(strides[0]) == len(strides) and padding.count(padding[0]) == len(padding) and pool_size[0] == padding[0]*2 + 1:
                kernel_shape_str = ', '.join('%s' % i for i in IR_node.get_attr('kernel_shape'))
                strides_str = ', '.join('%s' % i for i in IR_node.get_attr('strides'))

                code = "{:<15} = tf.nn.{}{}({}, [{}], [{}], padding='{}', name='{}')".format(
                    IR_node.variable_name,
                    op,
                    dim_str,
                    self.parent_variable_name(IR_node),
                    kernel_shape_str,
                    strides_str,
                    'SAME',
                    IR_node.name)
            else:
                kernel_shape_str = ', '.join('%s' % i for i in IR_node.get_attr('kernel_shape'))
                strides_str = ', '.join('%s' % i for i in IR_node.get_attr('strides'))
                input_node, padding = self._defuse_padding(IR_node, padding_const)
                code = "{:<15} = tf.nn.{}{}({}, [{}], [{}], padding='{}', name='{}')".format(
                    IR_node.variable_name,
                    op,
                    dim_str,
                    input_node,
                    kernel_shape_str,
                    strides_str,
                    padding,
                    IR_node.name)

        return code

    def emit_UNKNOWN(self, IR_node):
        print(IR_node.name)


    def emit_Add(self, IR_node):
        code = "{:<15} = {}".format(
            IR_node.variable_name,
            ' + '.join('%s' % self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))))
        
        return code

    def emit_DataInput(self, IR_node):
        assert not IR_node.in_edges
        shape_str = self._shapeToStr(IR_node.layer.attr["shape"].shape)

        if 'dtype' in IR_node.layer.attr:
            dtype_str = "{}, ".format(self.dtype_map[IR_node.layer.attr['dtype'].type])
        else:
            dtype_str = "tf.float32,"

        code = "{:<15} = tf.placeholder({} shape = ({}), name = '{}')".format(
            IR_node.variable_name, dtype_str, shape_str, IR_node.name
        )
        return code

    def emit_Dropout(self, IR_node):
        parent = self.IR_graph.get_parent(IR_node.name, [0])
        if self.trainable:
            self.add_body(1, "{:<15} = Dropout(name = '{}', dropout_rate = {})({})".format(
                IR_node.variable_name,
                IR_node.name,
                1 - IR_node.IR_layer.attr["keep_prob"].f,
                parent.real_variable_name))
        else:
            IR_node.real_name = parent.real_name


    def emit_FullyConnected(self, IR_node):
        if IR_node.name in self.weights_dict and 'weights' in self.weights_dict[IR_node.name]:
            kernel_str = "kernel_initializer = tf.constant_initializer(__weights_dict['{}']['weights']), ".format(IR_node.name)
        else: kernel_str = ""

        if IR_node.name in self.weights_dict and 'bias' in self.weights_dict[IR_node.name]:
            bias_str = "bias_initializer = tf.constant_initializer(__weights_dict['{}']['bias']), ".format(IR_node.name)
        else: bias_str = ""

        # check whether flatten operator should be added
        parent = self.IR_graph.get_parent(IR_node.name, [0])
        parent_shape = shape_to_list(parent.get_attr('_output_shapes')[0])
        if len(parent_shape) > 2:
            # flatten is needed
            self.add_body(1, "{:<15} = tf.contrib.layers.flatten({})".format(
                IR_node.variable_name + '_flatten',
                self.parent_variable_name(IR_node)))

            code = "{:<15} = tf.layers.dense({}, {}, {}{}use_bias = {})".format(
                IR_node.variable_name,
                IR_node.variable_name + '_flatten',
                IR_node.layer.attr['units'].i,
                kernel_str,
                bias_str,
                IR_node.layer.attr['use_bias'].b)
            return code

        else:
            code = "{:<15} = tf.layers.dense({}, {}, {}{}use_bias = {})".format(
                IR_node.variable_name,
                self.parent_variable_name(IR_node),
                IR_node.layer.attr['units'].i,
                kernel_str,
                bias_str,
                IR_node.layer.attr['use_bias'].b)
            return code


    def emit_UpSampling2D(self, IR_node):
        scales = IR_node.get_attr('scales')
        scales = tuple(scales)

        code = "{:<15} = tf.keras.layers.UpSampling2D(size={})({})".format(
            IR_node.variable_name,
            scales,
            self.parent_variable_name(IR_node))
        return code

    def emit_Flatten(self, IR_node):
        #self._emit_unary_operation(IR_node, "contrib.layers.flatten")
        code = "{:<15} = tf.contrib.layers.flatten({})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node))
        return code


    def emit_Mul(self, IR_node):

        code = "{:<15} = {}".format(
            IR_node.variable_name,
            ' * '.join('%s' % self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))))
        return code


    def emit_Const(self, IR_node):
        if 'dtype' in IR_node.layer.attr:
            dtype_str = "dtype={}".format(self.dtype_map[IR_node.layer.attr['dtype'].type])
            if 'int' in dtype_str:
                code = "{:<15} = tf.constant({}, {}, shape=(1,))".format(
                    IR_node.variable_name,
                    IR_node.layer.attr['value'].i,
                    dtype_str)
            else:
                code = "{:<15} = tf.constant({}, {}, shape=(1,))".format(
                    IR_node.variable_name,
                    IR_node.layer.attr['value'].f,
                    dtype_str)
        else:
            dtype_str = "dtype=tf.float32"
            code ="{:<15} = tf.constant({}, {}, shape=(1,))".format(
                IR_node.variable_name,
                IR_node.layer.attr['value'].f,
                dtype_str)

        return code

    def emit_Transpose(self, IR_node):
        code ="{:<15} = tf.transpose(a = {}, perm = {})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node, [0]),
            self.parent_variable_name(IR_node, [1]))
        
        return code

    def emit_Gather(self, IR_node):
        variable_str = "tf.convert_to_tensor(__weights_dict['{}']['weights'])".format(IR_node.name)

        code = "{:<15} = tf.gather(params = {}, indices = {}, axis = {})".format(
            IR_node.variable_name,
            variable_str,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('axis')
            )
        
        return code

    def emit_Unstack(self, IR_node):
        code = "{:<15} = tf.unstack(value={}, num={}, axis={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('num'),
            IR_node.get_attr('axis')
        )
        return code

    def emit_Reshape(self, IR_node):
        code = "{:<15} = tf.reshape({}, [{}], '{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            ', '.join('%s' % i for i in IR_node.get_attr('shape')),
            IR_node.name)
        
        return code


    def emit_Sub(self, IR_node):
        code = "{:<15} = {}".format(
            IR_node.variable_name,
            ' - '.join('%s' % self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))))
        
        return code

    def emit_Div(self, IR_node):
        code = "{:<15} = tf.div({}, {}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]),
            IR_node.name
        )
        return code

    def _emit_unary_operation(self, IR_node, op_name):
        code = "{:<15} = tf.{}({}, name = '{}')".format(
            IR_node.variable_name,
            op_name,
            self.parent_variable_name(IR_node),
            IR_node.name)
        return code

    def emit_Tanh(self, IR_node):
        code = self._emit_unary_operation(IR_node, 'tanh')
        return code

    def emit_Elu(self, IR_node):
        return self._emit_unary_operation(IR_node, 'nn.elu')


    def emit_Relu(self, IR_node):
        return self._emit_unary_operation(IR_node, 'nn.relu')


    def emit_Relu6(self, IR_node):
        return self._emit_unary_operation(IR_node, 'nn.relu6')


    def emit_CRelu(self, IR_node):
        return self._emit_unary_operation(IR_node, 'nn.crelu')


    def emit_PRelu(self, IR_node):
        self.used_layers.add(IR_node.type)
        code = "{:<15} = prelu({}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.name)
        return code

    def emit_LeakyRelu(self, IR_node):
        self.add_body(1, "{:<15} = tf.nn.leaky_relu({}, alpha={}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('alpha'),
            IR_node.name
        ))


    def emit_Softmax(self, IR_node):
        return self._emit_unary_operation(IR_node, 'nn.softmax')


    def emit_Sigmoid(self, IR_node):
        code = self._emit_unary_operation(IR_node, 'sigmoid')
        return code

    def emit_Embedding(self, IR_node):
        variable_str = "tf.convert_to_tensor(__weights_dict['{}']['weights'])".format(IR_node.name)
        code = "{:<15} = tf.nn.embedding_lookup(params = {}, ids = {})".format(
            IR_node.variable_name,
            variable_str,
            self.parent_variable_name(IR_node))
        return code

    def emit_LSTM(self, IR_node):
        return self.emit_RNNs(IR_node, "LSTM")


    def emit_GRU(self, IR_node):
        return self.emit_RNNs(IR_node, "GRU")


    def emit_Concat(self, IR_node):
        
        code = "{:<15} = tf.concat([{}], {}, name = '{}')".format(
            IR_node.variable_name,
            ', '.join(self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))),
            IR_node.layer.attr['axis'].i,
            IR_node.name)

        return code

    def emit_BatchNorm(self, IR_node):
        self.used_layers.add(IR_node.type)
        code = "{:<15} = batch_normalization({}, variance_epsilon={}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('epsilon'),
            IR_node.name)
        return code

    def emit_Scale(self, IR_node):
        self.used_layers.add(IR_node.type)
        code = "{:<15} = scale({}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.name)
        return code

    def emit_Pad(self, IR_node):
        padding = IR_node.get_attr('pads')
        padding = convert_onnx_pad_to_tf(padding)

        mode = IR_node.get_attr('mode', 'constant')
        mode = mode.lower()
        if mode == 'constant' or mode == 'reflect':
            mode = mode.upper()
        elif mode == 'edge':
            mode = 'SYMMETRIC'
        else:
            raise NotImplementedError("Not support padding mode {}.".format(mode))
        code = "{:<15} = tf.pad({}, {}, '{}', name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            padding,
            mode,
            IR_node.variable_name)
        return code

    def emit_Squeeze(self, IR_node):
        code = "{:<15} = tf.squeeze({}, [{}], name = '{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            ', '.join('%s' % axis for axis in IR_node.layer.attr['axes'].list.i),
            IR_node.name)
        return code


    def emit_ReduceMean(self, IR_node):
        code = "{:<15} = tf.reduce_mean({}, [{}], {}, name = '{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            ','.join('%s' % i for i in IR_node.get_attr('axes')),
            IR_node.get_attr('keepdims'),
            IR_node.name)
        return code

    def emit_LRN(self, IR_node):
        code = "{:<15} = tf.nn.lrn({}, depth_radius={}, bias={}, alpha={}, beta={}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('size') - 1,
            IR_node.get_attr('bias', 1),
            IR_node.get_attr('alpha') / (IR_node.get_attr('size') * 2 - 1),
            IR_node.get_attr('beta'),
            IR_node.name)
        return code

    def emit_SeparableConv(self, IR_node):
        self.used_layers.add(IR_node.type)
        strides_str = ', '.join('%s' % i for i in IR_node.get_attr('strides'))
        input_node, padding = self._defuse_padding(IR_node)
        code = "{:<15} = separable_convolution({}, strides = [{}], padding = '{}', name = '{}')".format(
            IR_node.variable_name,
            input_node,
            strides_str,
            padding,
            IR_node.name)
        return code


    def emit_DepthwiseConv(self, IR_node):
        self.used_layers.add(IR_node.type)
        strides_str = ', '.join('%s' % i for i in IR_node.layer.attr['strides'].list.i)
        input_node, padding = self._defuse_padding(IR_node)
        code = "{:<15} = depthwise_convolution({}, strides = [{}], padding = '{}', name = '{}')".format(
            IR_node.variable_name,
            input_node,
            strides_str,
            padding,
            IR_node.name)
        return code

    def emit_Crop(self, IR_node):
        border = IR_node.get_attr('border')
        assert len(border) == 4

        output_shape = IR_node.get_attr('_output_shapes')[0]
        output_shape = shape_to_list(output_shape)

        code = "{:<15} = tf.image.crop_to_bounding_box({}, offset_height={}, offset_width={}, target_height={}, target_width={})".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            border[0],
            border[1],
            output_shape[1],
            output_shape[2])
        
        return code

    def emit_ConvTranspose(self, IR_node):
        self.used_layers.add(IR_node.type)
        output_shape = [1] + shape_to_list(IR_node.get_attr('_output_shapes')[0])[1:]
        input_node, padding = self._defuse_padding(IR_node)
        code = "{:<15} = convolution_transpose({}, output_shape={}, strides={}, padding='{}', name='{}')".format(
            IR_node.variable_name,
            input_node,
            output_shape,
            IR_node.get_attr('strides'),
            padding,
            IR_node.name)
        return code

    def emit_Slice(self, IR_node):
        extra_str = ""
        if IR_node.get_attr('begin_mask'):
            extra_str += ", begin_mask={}".format(IR_node.get_attr('begin_mask'))
        if IR_node.get_attr('end_mask') != None:
            extra_str += ", end_mask={}".format(IR_node.get_attr('end_mask'))
        if IR_node.get_attr('shrink_axis_mask') != None:
            extra_str += ", shrink_axis_mask={}".format(IR_node.get_attr('shrink_axis_mask'))
        if IR_node.get_attr('new_axis_mask')!= None:
            extra_str += ", new_axis_mask={}".format(IR_node.get_attr('new_axis_mask'))

        if IR_node.get_attr('starts') != None:
            starts = IR_node.get_attr('starts')
        else:
            starts = self.parent_variable_name(IR_node, [1])
        
        if IR_node.get_attr('ends') != None:
            ends = IR_node.get_attr('ends')
        else:
            ends = self.parent_variable_name(IR_node, [2])
        
        if IR_node.get_attr('strides') != None:
            strides = IR_node.get_attr('strides')
        else:
            strides = self.parent_variable_name(IR_node, [3])

        code = "{:<15} = tf.strided_slice({}, {}, {}, {} {}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            starts,
            ends,
            strides,
            extra_str,
            IR_node.name)
        
        return code


    def emit_Shape(self, IR_node):
        code = "{:<15} = tf.shape({}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.name)
        return code

    def emit_Pack(self, IR_node):
        code = "{:<15} = tf.stack({}, axis={}, name='{}')".format(
            IR_node.variable_name,
            '[' +  ','.join('%s' % self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))) + ']',
            IR_node.get_attr('axis'),
            IR_node.name)
        return code

    def emit_Split(self, IR_node):
        code = "{:<15} = tf.split({}, {}, {}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('split'),
            IR_node.get_attr('axis'),
            IR_node.name)
        return code

    def emit_Unsqueeze(self, IR_node):
        code = "{:<15} = tf.expand_dims({}, axis={}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('axes')[0],
            IR_node.name)
        return code

    def emit_Fill(self, IR_node):
        code = "{:<15} = tf.fill({}, {}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            IR_node.get_attr('value'),
            IR_node.name)
        return code

    def emit_Maxmum(self, IR_node):
        code = "{:<15} = tf.maxmum({}, {}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]),
            IR_node.name
        )
        return code

    def emit_Minimum(self, IR_node):
        code = "{:<15} = tf.minimum({}, {}, name='{}')".format(
            IR_node.variable_name,
            self.parent_variable_name(IR_node),
            self.parent_variable_name(IR_node, [1]),
            IR_node.name
        )
        return code

    def emit_Scope(self, IR_node):
        input_vars = [self.parent_variable_name(IR_node, [idx]) for idx in range(len(IR_node.in_edges))]
        input_vars.append('__weights_dict')
        code = "{:<15} = _{}({})".format(
            IR_node.real_variable_name,
            IR_node.pattern,
            ', '.join(input_vars))
        self._gen_scope_code(IR_node)
        return code


    def _gen_scope_code(self, scope_node):

        def _scope_func(scope_name, params, code, return_var):
            code = """
def _{}({}):
{}
    return {}
    """.format(scope_name, params, code, ', '.join(return_var))
            return code

        if not self.layers_codes.get(scope_node.pattern, None):
            body_code = str()
            for node_name in scope_node.topology_list:
                node = self.IR_graph.get_node(node_name)
                node_type = node.type

                if hasattr(self, "emit_" + node_type):
                    func = getattr(self, "emit_" + node_type)
                    line = func(node)
                    if line != None:
                        body_code += "    " + line + '\n'
                else:
                    print("TensorflowEmitter has not supported operator [%s]." % (node_type))
                    self.emit_UNKNOWN(node)

            # param_code does not need parameter slice.
            input_params = scope_node.input_params
            input_params.append("__weights_dict")
            param_code = ', '.join(input_params)
            function_code = _scope_func(scope_node.pattern, param_code, body_code, scope_node.return_variables)

            self.layers_codes[scope_node.pattern] = function_code



    def _layer_Conv(self):
        self.add_body(0, """
def convolution(input, name, group, **kwargs):
    w = tf.Variable(__weights_dict[name]['weights'], trainable=is_train, name=name + "_weight")
    if group == 1:
        layer = tf.nn.convolution(input, w, name=name, **kwargs)
    else:
        weight_groups = tf.split(w, num_or_size_splits=group, axis=-1)
        xs = tf.split(input, num_or_size_splits=group, axis=-1)
        convolved = [tf.nn.convolution(x, weight, name=name, **kwargs) for
                    (x, weight) in zip(xs, weight_groups)]
        layer = tf.concat(convolved, axis=-1)

    if 'bias' in __weights_dict[name]:
        b = tf.Variable(__weights_dict[name]['bias'], trainable=is_train, name=name + "_bias")
        layer = layer + b
    return layer""")


    def _layer_PRelu(self):
        self.add_body(0, """
def prelu(input, name):
    gamma = tf.Variable(__weights_dict[name]['gamma'], name=name + "_gamma", trainable=is_train)
    return tf.maximum(0.0, input) + gamma * tf.minimum(0.0, input)
    """)


    def _layer_BatchNorm(self):
        self.add_body(0, """
def batch_normalization(input, name, **kwargs):
    mean = tf.Variable(__weights_dict[name]['mean'], name = name + "_mean", trainable = is_train)
    variance = tf.Variable(__weights_dict[name]['var'], name = name + "_var", trainable = is_train)
    offset = tf.Variable(__weights_dict[name]['bias'], name = name + "_bias", trainable = is_train) if 'bias' in __weights_dict[name] else None
    scale = tf.Variable(__weights_dict[name]['scale'], name = name + "_scale", trainable = is_train) if 'scale' in __weights_dict[name] else None
    return tf.nn.batch_normalization(input, mean, variance, offset, scale, name = name, **kwargs)
""")


    def _layer_Scale(self):
        self.add_body(0, """
def scale(input, name, **kwargs):
    mean = tf.Variable(__weights_dict[name]['scale_mean'], name = name + "_mean", trainable = is_train)
    variance = tf.Variable(__weights_dict[name]['scale_var'], name = name + "_var", trainable = is_train)
    offset = tf.Variable(__weights_dict[name]['bias'], name = name + "_bias", trainable = is_train) if 'bias' in __weights_dict[name] else None
    scale = tf.Variable(__weights_dict[name]['scale'], name = name + "_scale", trainable = is_train) if 'scale' in __weights_dict[name] else None
    return tf.nn.batch_normalization(input, mean, variance, offset, scale, variance_epsilon = 0, name = name)
""")


    def _layer_SeparableConv(self):
        self.add_body(0, """
def separable_convolution(input, name, **kwargs):
    depthwise = tf.Variable(__weights_dict[name]['depthwise_filter'], trainable = is_train, name = name + "_df")
    pointwise = tf.Variable(__weights_dict[name]['pointwise_filter'], trainable = is_train, name = name + "_pf")
    layer = tf.nn.separable_conv2d(input, depthwise, pointwise, **kwargs)
    if 'bias' in __weights_dict[name]:
        b = tf.Variable(__weights_dict[name]['bias'], trainable = is_train, name = name + "_bias")
        layer = layer + b
    return layer""")


    def _layer_DepthwiseConv(self):
        self.add_body(0, """
def depthwise_convolution(input, name, **kwargs):
    depthwise = tf.Variable(__weights_dict[name]['weights'], trainable = is_train, name = name + "_df")
    layer = tf.nn.depthwise_conv2d(input, depthwise, **kwargs)
    if 'bias' in __weights_dict[name]:
        b = tf.Variable(__weights_dict[name]['bias'], trainable = is_train, name = name + "_bias")
        layer = layer + b
    return layer""")


    def _layer_ConvTranspose(self):
        self.add_body(0, """
def convolution_transpose(input, name, **kwargs):
    w = tf.Variable(__weights_dict[name]['weights'], trainable=is_train, name=name + "_weight")
    dim = __weights_dict[name]['weights'].ndim - 2
    if dim == 2:
        layer = tf.nn.conv2d_transpose(input, w, **kwargs)
    elif dim == 3:
        layer = tf.nn.conv3d_transpose(input, w, **kwargs)
    else:
        raise ValueError("Error dim number {} in ConvTranspose".format(dim))

    if 'bias' in __weights_dict[name]:
        b = tf.Variable(__weights_dict[name]['bias'], trainable=is_train, name=name + "_bias")
        layer = layer + b
    return layer""")



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------
from six import string_types as _string_types

import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType


class Emitter(object):

    def __init__(self):
        self.body_code = str()
        self.weights_dict = dict()
        self.used_layers = set()
        self.weight_loaded = False
        self.layers_codes = dict()


    def run(self, dstNetworkPath, dstWeightPath = None, phase = 'test'):
        self.save_code(dstNetworkPath, phase)


    # share functions
    def add_body(self, indent, codes):
        if isinstance(codes, _string_types):
            codes = [codes]
        for code in codes:
            self.body_code += ("    " * indent) + code + '\n'

    def _load_weights(self, file_name=None):
        import numpy as np
        self.weight_loaded = True
        try:
            self.weights_dict = np.load(file_name).item()
        except:
            self.weights_dict = np.load(file_name, encoding='bytes').item()


    def parent_variable_name(self, IR_node, path_or_name = [0]):
        if isinstance(path_or_name, _string_types):
            path = [IR_node.in_edges.index(path_or_name)]
        elif isinstance(path_or_name, list):
            path = path_or_name
        else:
            raise ValueError
        return self.IR_graph.get_parent_variable_name(IR_node.name, path)

    def _build(self):
        self.IR_graph.build()


    def gen_code(self, phase):
        raise NotImplementedError("do not use base emitter class.")


    def save_code(self, filepath, phase):
        code = self.gen_code(phase)
        with open(filepath, 'w') as fout:
            fout.write(code)
        print("Target network code snippet is saved as [{}].".format(filepath))


    @staticmethod
    def save_weights(weights, filename):
        import numpy as np
        with open(filename, 'wb') as of:
            np.save(of, weights)
        print("Target weights are saved as [{}].".format(filename))



    @staticmethod
    def _image_in_transpose_str(dim):
        dims = [dim]
        dims.extend(range(dim))
        return ','.join('%s' % id for id in dims)


    @staticmethod
    def _image_out_transpose_str(dim):
        dims = list(range(1, dim + 1))
        dims.append(0)
        return ','.join('%s' % id for id in dims)


    @staticmethod
    def _conv_kernel_transpose_str(dim):
        dims = [dim + 1, dim]
        dims.extend(range(dim))
        return ','.join('%s' % id for id in dims)


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import six
from six import string_types as _string_types
from mmdnn.conversion.caffe.errors import ConversionError
from mmdnn.conversion.common.IR.graph_pb2 import GraphDef, NodeDef, TensorShape
from mmdnn.conversion.caffe.utils import get_real_name


def assign_attr_value(attr, val):
    '''Assign value to AttrValue proto according to data type.'''
    if isinstance(val, bool):
        attr.b = val
    elif isinstance(val, six.integer_types):
        attr.i = val
    elif isinstance(val, float):
        attr.f = val
    elif isinstance(val, str):
        attr.s = val.encode('utf-8')
    elif isinstance(val, TensorShape):
        attr.shape.MergeFromString(val.SerializeToString())
    elif isinstance(val, list):
        if len(val) == 0: return

        if isinstance(val[0], six.integer_types):
            attr.list.i.extend(val)
        elif isinstance(val[0], TensorShape):
            attr.list.shape.extend(val)
        else:
            raise NotImplementedError('AttrValue cannot be of %s %s' % (type(val), type(val[0])))
    else:
        raise NotImplementedError('AttrValue cannot be of %s' % type(val))


def fetch_attr_value(attr):
    '''Fetch valid value from AttrValue proto.'''
    field = attr.WhichOneof('value')
    val = getattr(attr, field) if field else None
    return val.decode('utf-8') if isinstance(val, bytes) else val


class Node(object):
    '''An intermediate representation for DL operations.'''

    def __init__(self, node_pb2):
        assert isinstance(node_pb2, NodeDef)
        self.node_pb2 = node_pb2
        self.output = []

    @staticmethod
    def create(op, **kwargs):
        node_pb2 = NodeDef()
        node_pb2.op = op
        for k, v in kwargs.items():
            assign_attr_value(node_pb2.attr[k], v)
        return Node(node_pb2)

    @property
    def op(self):
        return self.node_pb2.op

    @property
    def name(self):
        return self.node_pb2.name

    @name.setter
    def name(self, value):
        assert isinstance(value, _string_types)
        self.node_pb2.name = value

    @property
    def input(self):
        return self.node_pb2.input

    @property
    def attr(self):
        return self.node_pb2.attr.items()


class Graph(object):
    '''An intermediate representation for DL graph.'''

    def __init__(self, name, node_list, version=0):
        if node_list and len(node_list):
            assert isinstance(node_list[0], Node)
            self.node_dict = {node.name: node for node in node_list}
        else:
            self.node_dict = {}
        self.name = name
        self.version = version

    def topologically_sorted(self):
        visited = set()
        sorted_nodes = []
        def topo_sort_dfs(node, visited, sorted_nodes):
            if node in visited:
                return
            visited.add(node)
            for n in self.get_input(node):
                topo_sort_dfs(n, visited, sorted_nodes)
            sorted_nodes.append(node)
        for node in self.node_dict.values():
            topo_sort_dfs(node, visited, sorted_nodes)
        return sorted_nodes

    def get_node(self, name):
        return self.node_dict[name]

    def add_node(self, node):
        assert node.name not in self.node_dict
        self.node_dict[node.name] = node

    def remove_node(self, name):
        return self.node_dict.pop(name)

    def get_input(self, node):
        input_nodes = []
        for name in node.input:
            name = get_real_name(name)
            if name in self.node_dict:
                input_nodes.append(self.get_node(name))
        return input_nodes

    def as_graph_def(self):
        graph_pb2 = GraphDef()
        graph_pb2.version = self.version
        graph_pb2.node.extend([node.node_pb2 for node in self.node_dict.values()])
        return graph_pb2


#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import os
import sys
import math
import mxnet as mx
import numpy as np
from mmdnn.conversion.mxnet.mxnet_graph import MXNetGraph
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.DataStructure.parser import Parser
from mmdnn.conversion.common.utils import *

class MXNetParser(Parser):

    dtype_map = {
        "int8"      : graph_pb2.DT_INT8,
        "int16"     : graph_pb2.DT_INT16,
        "int32"     : graph_pb2.DT_INT32,
        "int64"     : graph_pb2.DT_INT64,
        "uint8"     : graph_pb2.DT_UINT8,
        "uint16"    : graph_pb2.DT_UINT16,
        "uint32"    : graph_pb2.DT_UINT32,
        "uint64"    : graph_pb2.DT_UINT64,
        "float16"   : graph_pb2.DT_FLOAT16,
        "float32"   : graph_pb2.DT_FLOAT32,
        "float64"   : graph_pb2.DT_FLOAT64
    }

    activation_map = {
        "relu"      : "Relu",
        "sigmoid"   : "Sigmoid",
        "tanh"      : "Tanh",
        # Not support yet
        # "softrelu"  : "SoftReLU"
    }

    channels_last = ['NDHWC', 'NHWC', 'NWC']
    channels_first = ['NCDHW', 'NCHW', 'NCW']

    @property
    def src_graph(self):
        return self.mxnet_graph

    @staticmethod
    def str2bool(v):
        return v.lower() in ("1", "true")


    @staticmethod
    def str2intList(v):
        v = v.replace("(", "")
        v = v.replace(")", "")
        if v == "":
            return list()
        else:
            return [int(s) for s in v.split(',')]


    @staticmethod
    def transpose(data, dim):
        if dim == 1:
            data = data.transpose((2, 1, 0))
        elif dim == 2:
            data = data.transpose((2, 3, 1, 0))
        elif dim == 3:
            data = data.transpose((2, 3, 4, 1, 0))
        else:
            print("Warning: The weight of dim {0} cannot transpose" % dim)

        return data


    @staticmethod
    def _convert_axis(IR_node, axis):
        ndim = len(IR_node.attr['_output_shapes'].list.shape[0].dim)
        if axis == 0:
            return 0
        elif axis == 1:
            return ndim - 1
        else:
            return axis - 1


    def trace_shape(self, source_node, IR_node):
        input_node = self.IR_layer_map[IR_node.input[0]]
        while len(input_node.attr['_output_shapes'].list.shape[0].dim) <= 2:
            IR_node = input_node
            input_node = self.IR_layer_map[IR_node.input[0]]

        input_shape = list()
        for e in input_node.attr["_output_shapes"].list.shape[0].dim:
            input_shape.append(e.size)
        C = input_shape.pop()
        ret = [C] + input_shape[1:]
        return ret


    def check_pad_mode(self, source_node, IR_node):
        kernel = MXNetParser.str2intList(source_node.get_attr("kernel"))
        dim = len(kernel)

        pad = source_node.get_attr("pad", "()")
        if pad == "()":
            pad = list([0] * dim)
        else:
            pad = MXNetParser.str2intList(pad)

        stride = source_node.get_attr("stride")
        if stride == None:
            stride = list([1] * dim)
        else:
            stride = MXNetParser.str2intList(stride)

        dilate = source_node.get_attr("dilate")
        if dilate == None:
            dilate = list([1] * dim)
        else:
            dilate = MXNetParser.str2intList(dilate)

        input_shape = list()
        if len(source_node.in_edges) == 0 or IR_node.input[0] not in self.IR_layer_map:
            input_shape = self.data_shape
        else:
            for e in self.IR_layer_map[IR_node.input[0]].attr["_output_shapes"].list.shape[0].dim:
                input_shape.append(e.size)

        valid_flag = True
        same_flag = True

        for i in range(dim):
            if not pad[i] == 0:
                valid_flag = False
            output_shape = int(math.floor(float(input_shape[i] + 2 * pad[i] - dilate[i] * (kernel[i] - 1) - 1) / float(stride[i])) + 1)
            same_pad_shape = int(math.ceil(float(input_shape[i]) / float(stride[i])))
            if not output_shape == same_pad_shape:
                same_flag = False

        if valid_flag:
            return "VALID"
        elif same_flag:
            return "SAME"
        else:
            return "None"


    @staticmethod
    def _load_model(weights, epoch):
        """Load a mxnet model from disk

        Parameters
        ----------
        model_path: str
            Path where the model network/params path is (json/params file)

        prefix: str
            prefix for json file, e.g. prefix-symbol.json

        epoch: int
            save epoch number

        Returns
        -------
        model: A mxnet model
        params: A pair of dictionaries each mapping parameter names to NDArray values
        """

        # Load the model network and weights
        sym, arg_params, aux_params = mx.model.load_checkpoint(weights, int(epoch))

        # digraph = mx.viz.plot_network(sym, save_format='jpg') # For debugging
        # digraph.render()

        model = mx.mod.Module(symbol = sym)
        arg_params.update(aux_params)
        return model, arg_params

        '''
        MXNet new api does not support load data without data_shapes
        '''
        # model.bind(data_shapes = data_shapes)
        # model.init_params()

        # mod.load(model_path, epoch_num)
        # return mod.get_params()


    @staticmethod
    def _load_json_file(model_path):
        """Load a mxnet network json file

        Parameters
        ----------
        model_path: str
            Path where the model network/params path is (json/params file)

        (Deleted)
        prefix: str
            prefix for json file, e.g. prefix-symbol.json

        Returns
        -------
        data["nodes"]: all the layer information(including weights, bias) with format
            data["nodes"][layer_num][params = {"name", "op", "attr", "inputs"}]

        """
        import json

        # load the model network
        with open(model_path, 'r') as data_file:
            data = json.load(data_file)

        # adjust the data format
        assert isinstance(data["nodes"], list)
        return data["nodes"]


    def __init__(self, input_arg):

        super(MXNetParser, self).__init__()

        json_data = list()
        self.data_shape = tuple()
        # load model files into MXNet graph
        # data_shape arguments added to calculate infer_shape(required)
        # if isinstance(input_arg, basestring):
        if len(input_arg) == 2:
            with open(input_arg[0], 'r') as input_json:
                json_string = input_json.read()
                symbol = mx.sym.load_json(json_string)
                self.model = mx.mod.Module(symbol = symbol)
            json_data = MXNetParser._load_json_file(input_arg[0])
            self.data_shape = tuple([1] + list(map(int, input_arg[1])))

        elif len(input_arg) == 4:
            self.model, self.weight_data = MXNetParser._load_model(input_arg[1], input_arg[2])
            json_data = MXNetParser._load_json_file(input_arg[0])
            self.weight_loaded = True
            assert isinstance(input_arg[3], list)
            self.data_shape = tuple([1] + list(map(int, input_arg[3])))

        else:
            raise ValueError("the # of input arguments [{}] is not supported" % len(input_arg))

        # Build network graph
        self.data_format = 'None'
        self.mxnet_graph = MXNetGraph(self.model)
        self.mxnet_graph.build(json_data)


    def gen_IR(self):
        self.IR_layer_map = dict()
        for layer in self.mxnet_graph.topological_sort:
            current_node = self.mxnet_graph.get_node(layer)
            node_type = current_node.type

            if hasattr(self, "rename_" + node_type):
                func = getattr(self, "rename_" + node_type)
                func(current_node)

            else:
                self.rename_UNKNOWN(current_node)


    def _copy_and_reop(self, source_node, IR_node, new_op = None):
        new_op = source_node.type if new_op == None else new_op
        if source_node.name.startswith('_'):
            source_node.real_name = source_node.name[1:]
        IR_node.name = source_node.real_name
        IR_node.op = new_op
        self.IR_layer_map[IR_node.name] = IR_node


    def set_output_shape(self, source_node, IR_node):
        sym_group = self.model.symbol.get_internals()
        for sym in sym_group:
            if source_node.name == sym.name:
                arg_shape, output_shape, aux_shape = sym.infer_shape(data = self.data_shape)
                for idx in range(len(output_shape)):
                    output_list = list(output_shape[idx])

                    # transpose to channel last
                    if not self.data_format in MXNetParser.channels_last:
                        channel = output_list.pop(1)
                        output_list.append(channel)

                    if IR_node.op == "DataInput":
                        MXNetParser._copy_shape(IR_node, [-1] + output_list[1:])

                    shape = graph_pb2.TensorShape()
                    for dim in output_list:
                        new_dim = shape.dim.add()
                        if dim == None:
                            new_dim.size = -1
                        else:
                            new_dim.size = dim

                    IR_node.attr["_output_shapes"].list.shape.extend([shape])
                break

    def _convert_identity_operation(self, source_node, new_op=None):
        IR_node = self.IR_graph.node.add()

        # name, op
        self._copy_and_reop(source_node, IR_node, new_op)

        # input edge
        self.convert_inedge(source_node, IR_node)

        # output shape
        self.set_output_shape(source_node, IR_node)

        return IR_node

    def _defuse_padding(self, source_node):
        IR_node = self.IR_graph.node.add()
        IR_node.name = source_node.name + "_pad"
        IR_node.op = "Pad"
        # input edge
        self.convert_inedge(source_node, IR_node)

        self.IR_layer_map[IR_node.name] = IR_node

        # attr
        assign_IRnode_values(IR_node, {'mode' : 'CONSTANT'})
        # print("Warning: MXNet symbol pad does not support channel last")

        pad = MXNetParser.str2intList(source_node.get_attr("pad"))
        args['pads'] = [0, 0]
        for e in pad:
            args['pads'].extend([e, e])
        args['pads'] += [0, 0]
        args['pads'] = convert_tf_pad_to_onnx(args['pads'])
        IR_node.set_attrs(args)

        # IR_node.attr["pads"].list.i.extend([0, 0])
        # for e in pad:
        #     IR_node.attr["pads"].list.i.extend([e, e])
        # IR_node.attr["pads"].list.i.extend([0, 0])

        IR_node.attr["constant_values"].f = 0.


    @staticmethod
    def _copy_shape(IR_node, output_list):
        if not output_list == None:
            for dim in output_list:
                new_dim = IR_node.attr["shape"].shape.dim.add()
                if dim == None:
                    new_dim.size = -1
                else:
                    new_dim.size = dim
        else:
            IR_node.attr["shape"].shape.unknown_rank = True


    def rename_UNKNOWN(self, source_node):
        print("Warning: MXNet Parser has not supported operator %s with name %s."
            % (source_node.type, source_node.name))
        if source_node.type == "null" and source_node.name != 'label':
            print("Warning: convert the null operator with name [%s] into input layer." % source_node.name)
            IR_node = self.IR_graph.node.add()

            # name, op
            self._copy_and_reop(source_node, IR_node, "DataInput")

            # input edge
            self.convert_inedge(source_node, IR_node)

            self.set_output_shape(source_node, IR_node)

        else:
            raise NotImplementedError()



    """
    Here start with Neural Network Symbol
    """

    def rename_Pad(self, source_node):
        IR_node = self._convert_identity_operation(source_node)
        kwargs = dict()
        pad = MXNetParser.str2intList(source_node.get_attr("pad_width"))
        pad += [pad.pop(2), pad.pop(3)]
        kwargs['pads'] = pad
        kwargs['pads'] = convert_tf_pad_to_onnx(kwargs['pads'])
        kwargs['mode'] = 'CONSTANT'
        assign_IRnode_values(IR_node, kwargs)
        IR_node.attr["constant_values"].f = 0.


    def rename_FullyConnected(self, source_node):
        IR_node = self._convert_identity_operation(source_node)

        # units
        IR_node.attr["units"].i = int(source_node.get_attr("num_hidden"))

        # use bias (no_bias default = False)
        IR_node.attr["use_bias"].b = not MXNetParser.str2bool(source_node.get_attr("no_bias", "False"))

        # weights
        if self.weight_loaded:
            if self.data_format == 'NM':
                self.set_weight(source_node.name, "weights", self.weight_data.get(source_node.name + "_weight").asnumpy().transpose((1, 0)))
            else:
                weight = self.weight_data.get(source_node.name + "_weight").asnumpy().transpose((1, 0))
                original_shape = weight.shape

                channel_first_list = self.trace_shape(source_node, IR_node)
                dim = len(channel_first_list) + 1
                weight = weight.reshape(channel_first_list + [original_shape[1]])
                assert dim > 2
                weight = weight.transpose(list(range(1, dim-1)) + [0, dim-1])
                weight = weight.reshape(original_shape)
                self.set_weight(source_node.name, "weights", weight)

            if IR_node.attr["use_bias"].b:
                self.set_weight(source_node.name, "bias", self.weight_data.get(source_node.name + "_bias").asnumpy())

        if not self.data_format == 'NM':
            # print("Warning: Layer [{}] has changed model data format from [{}] to [NM]".format(source_node.name, self.data_format))
            self.data_format = 'NM'


    def rename_Convolution(self, source_node):
        IR_node = self.IR_graph.node.add()

        # input edge
        self.convert_inedge(source_node, IR_node)

        # output shape
        self.set_output_shape(source_node, IR_node)

        dim = 0
        layout = 'None'

        # kernel_shape
        kernel = MXNetParser.str2intList(source_node.get_attr("kernel"))
        dim = len(kernel)
        IR_node.attr["kernel_shape"].list.i.extend(kernel)

        layout = source_node.get_attr("layout")
        if layout == None or layout == 'None':
            if dim == 1:
                layout = "NCW"
            elif dim == 2:
                layout = "NCHW"
            elif dim == 3:
                layout = "NCDHW"

        if not self.data_format == layout:
            # print("Warning: Layer [{}] has changed model data format from [{}] to [{}]".format(source_node.name, self.data_format, layout))
            self.data_format = layout

        # groups
        group = int(source_node.get_attr("num_group", "1"))
        IR_node.attr["group"].i = group
        in_channel = self.IR_layer_map[IR_node.input[0]].attr["_output_shapes"].list.shape[0].dim[-1].size

        if group == in_channel:
            self._copy_and_reop(source_node, IR_node, "DepthwiseConv")
        else:
            self._copy_and_reop(source_node, IR_node, "Conv")
        # in_channel = in_channel // group

        out_channel = int(source_node.get_attr("num_filter"))

        IR_node.attr["kernel_shape"].list.i.extend([in_channel, out_channel])

        # use_bias (no_bias default = False)
        IR_node.attr["use_bias"].b = not MXNetParser.str2bool(source_node.get_attr("no_bias", "False"))

        # strides
        strides = source_node.get_attr("stride")
        IR_node.attr["strides"].list.i.append(1)
        if not strides == None:
            IR_node.attr["strides"].list.i.extend(MXNetParser.str2intList(strides))
        else:
            IR_node.attr["strides"].list.i.extend([1] * dim)
        IR_node.attr["strides"].list.i.append(1)

        # dilations
        dilate = source_node.get_attr("dilate")
        IR_node.attr["dilations"].list.i.append(1)
        if not dilate == None:
            IR_node.attr["dilations"].list.i.extend(MXNetParser.str2intList(dilate))
        else:
            IR_node.attr["dilations"].list.i.extend([1] * dim)
        IR_node.attr["dilations"].list.i.append(1)

        # data_format
        assign_IRnode_values(IR_node, {'data_format' : layout})

        # padding
        if "pad" in source_node.attr:
            pad = MXNetParser.str2intList(source_node.get_attr("pad"))
            IR_node.attr["pads"].list.i.extend(([0]+pad+[0])*2)
        else:
            IR_node.attr["pads"].list.i.extend([0, 0] * (dim + 2))

        # weights
        if self.weight_loaded:
            weight = self.weight_data.get(source_node.name + "_weight").asnumpy()
            if not layout in MXNetParser.channels_last:
                weight = MXNetParser.transpose(weight, dim)
                if IR_node.op == "DepthwiseConv":
                    weight = weight.transpose(0, 1, 3, 2)
            self.set_weight(source_node.name, "weights", weight)

            if IR_node.attr["use_bias"].b:
                self.set_weight(source_node.name, "bias", self.weight_data.get(source_node.name + "_bias").asnumpy())


    def rename_Activation(self, source_node):
        self._convert_identity_operation(source_node, new_op=MXNetParser.activation_map[source_node.get_attr("act_type")])


    def rename_BatchNorm(self, source_node):
        IR_node = self._convert_identity_operation(source_node)

        # axis
        if self.data_format in MXNetParser.channels_first or self.data_format == 'None':
            IR_node.attr["axis"].i = MXNetParser._convert_axis(IR_node, int(source_node.get_attr("axis", "1")))
        else:
            IR_node.attr["axis"].i = int(source_node.get_attr("axis", "1"))

        # scale
        IR_node.attr["scale"].b = not MXNetParser.str2bool(source_node.get_attr("fix_gamma", "True"))
        IR_node.attr["bias"].b = True
        # epsilon
        IR_node.attr["epsilon"].f = float(source_node.get_attr("eps", "0.001"))

        # momentum
        IR_node.attr["momentum"].f = float(source_node.get_attr("momentum", "0.9"))

        # weights
        if self.weight_loaded:
            # gamma
            if IR_node.attr["scale"].b:
                self.set_weight(source_node.name, "scale", self.weight_data.get(source_node.name + "_gamma").asnumpy())

            # beta
            if IR_node.attr["bias"].b:
                self.set_weight(source_node.name, "bias", self.weight_data.get(source_node.name + "_beta").asnumpy())

            # mean
            self.set_weight(source_node.name, "mean", self.weight_data.get(source_node.name + "_moving_mean").asnumpy())

            # var
            self.set_weight(source_node.name, "var", self.weight_data.get(source_node.name + "_moving_var").asnumpy())


    def rename_Pooling(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        self._copy_and_reop(source_node, IR_node, "Pool")

        # input edge
        self.convert_inedge(source_node, IR_node)

        # pooling type (sum not allowed yet)
        pool_type = source_node.get_attr("pool_type")
        if pool_type == "sum":
            print("Warning: sum pooling is not supported yet.")
        elif pool_type == "max":
            assign_IRnode_values(IR_node, {'pooling_type' : 'MAX'})
        elif pool_type == "avg":
            assign_IRnode_values(IR_node, {'pooling_type' : 'AVG'})
        else:
            raise ValueError("Error pool_type {}.".format(pool_type))

        kernel_shape = MXNetParser.str2intList(source_node.get_attr("kernel"))

        if MXNetParser.str2bool(source_node.get_attr("global_pool", "False")):

            IR_node.attr['global_pooling'].b = True
            IR_node.attr["kernel_shape"].list.i[:] = [1] * (len(kernel_shape) + 2)
            IR_node.attr["strides"].list.i[:] = [1] * (len(kernel_shape) + 2)
        else:
            IR_node.attr['global_pooling'].b = False

            # strides
            strides = source_node.get_attr("stride")
            IR_node.attr["strides"].list.i.append(1)
            if not strides == None:
                IR_node.attr["strides"].list.i.extend(MXNetParser.str2intList(strides))
            IR_node.attr["strides"].list.i.append(1)

            # kernel_shape
            IR_node.attr["kernel_shape"].list.i.append(1)
            IR_node.attr["kernel_shape"].list.i.extend(kernel_shape)
            IR_node.attr["kernel_shape"].list.i.append(1)

            # padding
            if "pad" in source_node.attr:
                pad = MXNetParser.str2intList(source_node.get_attr("pad"))
                IR_node.attr["pads"].list.i.extend(([0]+pad+[0])*2)
            else:
                IR_node.attr["pads"].list.i.extend(([0])*8)

        # output shape
        self.set_output_shape(source_node, IR_node)


    def rename_SoftmaxOutput(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        self._copy_and_reop(source_node, IR_node, "Softmax")

        # input edge
        self.convert_inedge(source_node, IR_node)

        if "attr" in source_node.layer or "param" in source_node.layer:
            print("Warning: SoftmaxOutput attrs are not supported in IR.")

        # output shape
        self.set_output_shape(source_node, IR_node)


    def rename_softmax(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Softmax')

        # dim
        if self.data_format in MXNetParser.channels_first or self.data_format == 'None':
            IR_node.attr["dim"].i = MXNetParser._convert_axis(IR_node, int(source_node.get_attr("axis", "-1")))
        else:
            IR_node.attr["dim"].i = int(source_node.get_attr("axis", "-1"))


    # def rename_log_softmax(self, source_node):
    #   raise NotImplementedError("not support yet")


    # def rename_Correlation(self, source_node):
    #   raise NotImplementedError("not support yet")


    def rename_Deconvolution(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        self._copy_and_reop(source_node, IR_node, "ConvTranspose")

        # input edge
        self.convert_inedge(source_node, IR_node)

        dim = 0
        layout = 'None'

        # padding
        if "pad" in source_node.attr:
            pad = MXNetParser.str2intList(source_node.get_attr("pad"))
            IR_node.attr["pads"].list.i.extend(([0]+pad+[0])*2)
        else:
            IR_node.attr["pads"].list.i.extend([0, 0] * (dim + 2))

        # output shape
        self.set_output_shape(source_node, IR_node)

        # kernel_shape
        kernel = MXNetParser.str2intList(source_node.get_attr("kernel"))
        dim = len(kernel)
        IR_node.attr["kernel_shape"].list.i.extend(kernel)

        layout = source_node.get_attr("layout")
        if layout == None or layout == 'None':
            if dim == 1:
                layout = "NCW"
            elif dim == 2:
                layout = "NCHW"
            elif dim == 3:
                layout = "NCDHW"

        if not self.data_format == layout:
            # print("Warning: Layer [{}] has changed model data format from [{}] to [{}]".format(source_node.name, self.data_format, layout))
            self.data_format = layout

        in_channel = self.IR_layer_map[IR_node.input[0]].attr["_output_shapes"].list.shape[0].dim[-1].size

        out_channel = int(source_node.get_attr("num_filter"))

        IR_node.attr["kernel_shape"].list.i.extend([out_channel, in_channel])

        # use_bias (no_bias default = False)
        IR_node.attr["use_bias"].b = not MXNetParser.str2bool(source_node.get_attr("no_bias", "False"))

        # strides
        strides = source_node.get_attr("strides")
        IR_node.attr["strides"].list.i.append(1)
        if not strides == None:
            IR_node.attr["strides"].list.i.extend(MXNetParser.str2intList(strides))
        else:
            IR_node.attr["strides"].list.i.extend([1] * dim)
        IR_node.attr["strides"].list.i.append(1)

        # dilations
        dilate = source_node.get_attr("dilate")
        IR_node.attr["dilations"].list.i.append(1)
        if not dilate == None:
            IR_node.attr["dilations"].list.i.extend(MXNetParser.str2intList(dilate))
        else:
            IR_node.attr["dilations"].list.i.extend([1] * dim)
        IR_node.attr["dilations"].list.i.append(1)

        # data_format
        IR_node.attr["data_format"].s = layout

        # groups
        IR_node.attr["group"].i = int(source_node.get_attr("num_group", "1"))

        # weights
        if self.weight_loaded:
            weight = self.weight_data.get(source_node.name + "_weight").asnumpy()
            if not layout in MXNetParser.channels_last:
                weight = MXNetParser.transpose(weight, dim)
            self.set_weight(source_node.name, "weights", weight)

            if IR_node.attr["use_bias"].b:
                self.set_weight(source_node.name, "bias", self.weight_data.get(source_node.name + "_bias").asnumpy())


    # def rename_RNN(self, source_node):
    #   raise NotImplementedError("RNN not support yet")


    def rename_Embedding(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        self._copy_and_reop(source_node, IR_node)

        # input edge
        self.convert_inedge(source_node, IR_node)

        # input_dim
        IR_node.attr["input_dim"].i = int(source_node.get_attr("input_dim"))

        # output_dim
        IR_node.attr["output_dim"].i = int(source_node.get_attr("output_dim"))

        # dtype
        IR_node.attr["dtype"].type = MXNetParser.dtype_map[source_node.get_attr("dtype", "float32")]

        # output shape
        self.set_output_shape(source_node, IR_node)


    # IR only support elu and prelu from {'elu', 'leaky', 'prelu', 'rrelu'}
    def rename_LeakyReLU(self, source_node):
        act_type = source_node.get_attr('act_type', None)
        if act_type:
            if not act_type == "elu" and not act_type == "prelu":
                print("Warning: Activation Type %s is not supported yet." % act_type)
                # return

        IR_node = self.IR_graph.node.add()

        # name, op
        if act_type == 'prelu':
            self._copy_and_reop(source_node, IR_node, "PRelu")

            # gamma
            self.set_weight(source_node.name, "gamma", self.weight_data.get(source_node.name + "_gamma").asnumpy())

        else:  # All other cases set to 'Elu'
            self._copy_and_reop(source_node, IR_node, "Elu")

        # input edge
        self.convert_inedge(source_node, IR_node)

        # alpha [exp(x) - alpha], but mxnet attr slope [slope*(exp(x) - 1)] when x < 0
        if "slope" in source_node.attr:
            raise ValueError("Attribute Slope is not supported in IR format")
        # IR_node.attr["alpha"].f = float()

        # output shape
        self.set_output_shape(source_node, IR_node)

        # raise NotImplementedError("slope cannot convert to alpha")


    # def rename_InstanceNorm(self, source_node):
    #   raise NotImplementedError


    # def rename_L2Normalization(self, source_node):
    #   raise NotImplementedError


    def rename_LRN(self, source_node):
        IR_node = self._convert_identity_operation(source_node)

        # alpha
        IR_node.attr["alpha"].f = float(source_node.get_attr("alpha", "0.0001"))
        # beta
        IR_node.attr["beta"].f = float(source_node.get_attr("beta", "0.75"))
        # knorm
        IR_node.attr["k"].f = float(source_node.get_attr("knorm", "2"))
        # nsize
        IR_node.attr["size"].i = float(source_node.get_attr["nsize"])


    def rename_ROIPooling(self, source_node):
        raise NotImplementedError()


    def rename_Dropout(self, source_node):
        IR_node = self._convert_identity_operation(source_node)

        # keep_prob
        IR_node.attr["keep_prob"].f = float(source_node.get_attr("p", "0.5"))

        # mode
        assign_IRnode_values(IR_node, {'mode' : 'training'})


    """
    Here start with Symbol manipulation routines
    """

    # reverse cannot support yet
    def rename_reshape(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Reshape')

        # old API target_shape not support yet
        shape = source_node.get_attr("shape")
        if not shape == None:
            shape_list = MXNetParser.str2intList(shape)
            for param in shape_list:
                if param <= 0 and not param == -1:
                    raise ValueError("special value %d for Reshape is not pre-defined in IR." % param)
            IR_node.attr["shape"].list.i.extend(shape_list)

        # output shape
        self.set_output_shape(source_node, IR_node)

        # raise NotImplementedError("adjust output shape")


    def rename_Flatten(self, source_node):
        self._convert_identity_operation(source_node, new_op='Flatten')


    def rename_Concat(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Concat')

        # dim
        if self.data_format in MXNetParser.channels_first or self.data_format == 'None':
            IR_node.attr["axis"].i = MXNetParser._convert_axis(IR_node, int(source_node.get_attr("dim", "1")))
        else:
            IR_node.attr["axis"].i = int(source_node.get_attr("dim", "1"))


    def rename_cast(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Cast')

        # dtype
        IR_node.attr["dtype"].type = MXNetParser.dtype_map[source_node.get_attr("dtype")]

        # output shape
        self.set_output_shape(source_node, IR_node)


    def rename_expand_dims(self, source_node):
        IR_node = self.IR_graph.node.add()

        # name, op
        self._copy_and_reop(source_node, IR_node)

        # input edge
        self.convert_inedge(source_node, IR_node)

        # output shape
        self.set_output_shape(source_node, IR_node)

        # axis
        if self.data_format in MXNetParser.channels_first or self.data_format == 'None':
            IR_node.attr["axis"].i = MXNetParser._convert_axis(IR_node, int(source_node.get_attr("axis")))
        else:
            IR_node.attr["axis"].i = int(source_node.get_attr("axis"))


    def rename_elemwise_add(self, source_node):
        self._convert_identity_operation(source_node, new_op='Add')


    def rename__Plus(self, source_node):
        self._convert_identity_operation(source_node, new_op='Add')


    def rename_broadcast_add(self, source_node):
        self._convert_identity_operation(source_node, new_op='Add')


    def rename_broadcast_mul(self, source_node):
        self._convert_identity_operation(source_node, new_op='Mul')


    def rename__mul(self, source_node):
        self._convert_identity_operation(source_node, new_op='Mul')


    def rename__copy(self, source_node):
        self._convert_identity_operation(source_node)
        # raise NotImplementedError("No matching IR api")


    def _convert_scalar_operator(self, source_node, new_op):
        value = source_node.get_attr('scalar')
        value_node = self.IR_graph.node.add()
        value_node.name = source_node.real_name + "_second"
        # left strip the "_" at the beginning of the name
        # Issue #85, #135
        value_node.name = value_node.name.lstrip('_')
        value_node.op = 'Constant'
        self.set_weight(value_node.name, 'value', np.array([value], np.float32))

        IR_node = self._convert_identity_operation(source_node, new_op)
        IR_node.input.append(value_node.name)
        return IR_node


    def rename__mul_scalar(self, source_node):
        self._convert_scalar_operator(source_node, 'Mul')


    def rename__minus_scalar(self, source_node):
        self._convert_scalar_operator(source_node, 'Sub')


    def rename__copy(self, source_node):
        source_node.real_name = self.get_parent(source_node.name, [0]).real_name


    def rename_BlockGrad(self, source_node):
        return




#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import os
import mxnet as mx
from mmdnn.conversion.common.DataStructure.graph import GraphNode, Graph

class MXNetGraphNode(GraphNode):



    def __init__(self, layer):
        super(MXNetGraphNode, self).__init__(layer)

        if "attr" in layer:
            self.attr = layer["attr"]
        elif "param" in layer:
            self.attr = layer["param"]
        elif "attrs" in layer:
            self.attr = layer["attrs"]
        else:
            self.attr = None


    @property
    def name(self):
        return self.layer["name"]


    @property
    def type(self):
        return self.layer["op"]


    @property
    def mx_layer(self):
        return self.layer


    def get_attr(self, name, default_value=None):
        if self.attr:
            if name in self.attr.keys():
                return self.attr.get(name)
            else:
                return default_value
        else:
            return default_value

class MXNetGraph(Graph):

    def __init__(self, model):
        # sanity check non-sense always input module.Module
        # if not (type(model) == mx.module.Module
        #     or type(model) == mx.module.SequentialModule
        #     or type(model) == mx.model)
        #     raise TypeError("MXNet layer of type %s is not supported." % type(model))

        super(MXNetGraph, self).__init__(model)


    def build(self, json_data):

        self.input_layers = list()
        input_dict = dict() # dict{layer_num, layer_name}
        layer_num = -1

        import re

        for layer in json_data:

            layer_num += 1
            # if layer["op"] == "null":
            #     continue

            if re.search("_(weight|bias|var|mean|gamma|beta|label)", layer["name"]) and layer["op"] == "null":
                continue

            input_dict.update({layer_num: layer["name"]})
            self.layer_map[layer["name"]] = MXNetGraphNode(layer)
            self.layer_name_map[layer["name"]] = layer["name"]
            for input_layer in layer["inputs"]:
                assert isinstance(input_layer, list)
                if input_layer[0] in input_dict:
                    pred = input_dict.get(input_layer[0])

                    if pred not in self.layer_map:
                        new_node = dict({'op': 'NoOp', 'name': pred, 'inputs': list()})
                        self.layer_map[pred] = MXNetGraphNode(new_node)
                        self.layer_name_map[pred] = pred

                    self._make_connection(pred, layer["name"])

        super(MXNetGraph, self).build()

        # raise NotImplementedError("Cannot support multi-input")


import sys as _sys
import google.protobuf.text_format as text_format
from six import text_type as _text_type



def _convert(args):
    if args.framework == 'caffe':
        raise NotImplementedError("Destination [Caffe] is not implemented yet.")

    elif args.framework == 'keras':
        raise NotImplementedError("Destination [Keras] is not implemented yet.")

    elif args.framework == 'tensorflow':
        raise NotImplementedError("Destination [Tensorflow] is not implemented yet.")

    elif args.framework == 'cntk':
        raise NotImplementedError("Destination [CNTK] is not implemented yet.")

    elif args.framework == 'coreml':
        from mmdnn.conversion.coreml.coreml_emitter import CoreMLEmitter
        assert args.inputNetwork is not None
        assert args.inputWeight is not None
        emitter = CoreMLEmitter(args.inputNetwork, args.inputWeight)
        model, in_, out_ = emitter.gen_model(
            args.inputNames,
            args.outputNames,
            image_input_names = set(args.imageInputNames) if args.imageInputNames else None,
            is_bgr = args.isBGR,
            red_bias = args.redBias,
            blue_bias = args.blueBias,
            green_bias = args.greenBias,
            gray_bias = args.grayBias,
            image_scale = args.scale,
            class_labels = args.classInputPath if args.classInputPath else None,
            predicted_feature_name = args.predictedFeatureName)

        """
        from google.protobuf import text_format
        with open(args.output+'.txt', 'w') as f:
            f.write(text_format.MessageToString(model))
        """

        with open(args.output, 'wb') as f:
            model = model.SerializeToString()
            f.write(model)


        return 0

    elif args.framework == 'pytorch':
        if not args.dstWeightPath or not args.IRWeightPath:
            raise ValueError("Need to set a target weight filename.")
        from mmdnn.conversion.pytorch.pytorch_emitter import PytorchEmitter
        emitter = PytorchEmitter((args.IRModelPath, args.IRWeightPath))

    elif args.framework == 'mxnet':
        from mmdnn.conversion.mxnet.mxnet_emitter import MXNetEmitter
        if args.IRWeightPath == None:
            emitter = MXNetEmitter(args.IRModelPath)
        else:
            emitter = MXNetEmitter((args.IRModelPath, args.IRWeightPath, args.inputShape, args.dstWeightPath))

    else:
        assert False

    emitter.run(args.output)

    return 0


def _get_parser():
    import argparse

    parser = argparse.ArgumentParser(description='Convert IR model file formats to other format.')

    parser.add_argument(
        '-f', '--framework', type=_text_type, choices=['coreml'], required=True,
        help='Format of model at srcModelPath (default is to auto-detect).'
    )

    parser.add_argument(
        '-in', '--inputNetwork',
        type=_text_type,
        required=True,
        help='Path of the IR network architecture file.')

    parser.add_argument(
        '-iw', '--inputWeight',
        type=_text_type,
        required=True,
        help='Path to the IR network weight file.')

    parser.add_argument(
        '-o', '--output',
        type=_text_type,
        required=True,
        help='Path to save the destination model')

    # Caffe
    parser.add_argument(
        '--phase', type=_text_type, choices=['train', 'test'], default='test',
        help='[Caffe] Convert phase (train/test) for destination toolkits.'
    )

    # For CoreML
    parser.add_argument('--inputNames', type=_text_type, nargs='*', help='Names of the feature (input) columns, in order (required for keras models).')
    parser.add_argument('--outputNames', type=_text_type, nargs='*', help='Names of the target (output) columns, in order (required for keras models).')
    parser.add_argument('--imageInputNames', type=_text_type, default=[], action='append', help='Label the named input as an image. Can be specified more than once for multiple image inputs.')
    parser.add_argument('--isBGR', action='store_true', default=False, help='True if the image data in BGR order (RGB default)')
    parser.add_argument('--redBias', type=float, default=0.0, help='Bias value to be added to the red channel (optional, default 0.0)')
    parser.add_argument('--blueBias', type=float, default=0.0, help='Bias value to be added to the blue channel (optional, default 0.0)')
    parser.add_argument('--greenBias', type=float, default=0.0, help='Bias value to be added to the green channel (optional, default 0.0)')
    parser.add_argument('--grayBias', type=float, default=0.0, help='Bias value to be added to the gray channel for Grayscale images (optional, default 0.0)')
    parser.add_argument('--scale', type=float, default=1.0, help='Value by which the image data must be scaled (optional, default 1.0)')
    parser.add_argument('--classInputPath', type=_text_type, default='', help='Path to class labels (ordered new line separated) for treating the neural network as a classifier')
    parser.add_argument('--predictedFeatureName', type=_text_type, default='class_output', help='Name of the output feature that captures the class name (for classifiers models).')
    return parser


def _main():
    parser=_get_parser()
    args = parser.parse_args()
    ret = _convert(args)
    _sys.exit(int(ret)) # cast to int or else the exit code is always 1


if __name__ == '__main__':
    _main()



#----------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License. See License.txt in the project root for license information.
#----------------------------------------------------------------------------------------------

import os
import numpy as np
from six.moves import xrange
import cntk as _cntk
from mmdnn.conversion.cntk.cntk_graph import CntkGraph
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.common.IR.graph_pb2 import NodeDef, GraphDef, DataType
from mmdnn.conversion.common.utils import *
from mmdnn.conversion.common.DataStructure.parser import Parser


class CntkParser(Parser):

    dtype_map = {
        0 : graph_pb2.DT_UNDEFINED,
        np.float32 : graph_pb2.DT_FLOAT32,
        np.float64 : graph_pb2.DT_FLOAT64,
        3 : graph_pb2.DT_INT32,
        4 : graph_pb2.DT_UINT8,
        5 : graph_pb2.DT_INT16,
        6 : graph_pb2.DT_INT8,
        7 : graph_pb2.DT_STRING,
        9 : graph_pb2.DT_INT64
    }


    @property
    def src_graph(self):
        return self.cntk_graph


    def __init__(self, model, dest_nodes = None):
        super(CntkParser, self).__init__()

        if not os.path.exists(model):
            raise ValueError('Cntk model [{}] can not be found!'.format(model))
        model = _cntk.Function.load(model)
        self.weight_loaded = True

        # Build network graph
        self.cntk_graph = CntkGraph(model)
        self.cntk_graph.build()


    @staticmethod
    def _convert_padding_to_IR(kernel_shape, auto_pad):
        lower = []
        upper = []
        for idx in range(0, len(kernel_shape)):
            if auto_pad[idx] == False:
                lower += [0]
                upper += [0]
            else:
                q = kernel_shape[idx] // 2
                lower += [q] if kernel_shape[idx] % 2 else [q - 1]
                upper += [q]

        return [0] + lower + [0, 0] + upper + [0]


    def _convert_identity_operation(self, source_node, start_edge=0, end_edge=None, new_op=None, shape_transpose=True):
        IR_node = self.IR_graph.node.add()
        CntkParser._copy_and_reop(source_node, IR_node, new_op, shape_transpose)
        self.convert_inedge(source_node, IR_node, start_edge, end_edge)
        return IR_node


    def gen_IR(self):
        for layer in self.src_graph.topological_sort:
            current_node = self.src_graph.get_node(layer)
            node_type = current_node.type
            if hasattr(self, "rename_" + node_type):
                func = getattr(self, "rename_" + node_type)
                func(current_node)
            else:
                self.rename_UNKNOWN(current_node)


    @staticmethod
    def _copy_and_reop(source_node, IR_node, new_op=None, shape_transpose=False):
        if new_op == None: new_op = source_node.type
        IR_node.name = source_node.real_name
        IR_node.op = new_op

        kwargs = {}

        if hasattr(source_node.layer, 'dtype'):
            assert source_node.layer.dtype in CntkParser.dtype_map, 'type [{}] is unknown.'.format(source_node.layer.dtype)
            IR_node.attr["dtype"].type = CntkParser.dtype_map[source_node.layer.dtype]

        if hasattr(source_node.layer, 'shape'):
            shape =  (-1,) + source_node.layer.shape
            if shape_transpose:
                shape = CntkParser.channel_first_shape_to_IR(shape)
            shape = list_to_shape(shape)
            kwargs['_output_shapes'] = [shape]

        assign_IRnode_values(IR_node, kwargs)


    def _fuse_bias_node(self, source_node):
        next_node = self.src_graph.get_son(source_node.name, [0])
        if next_node is None or next_node.type != 'Plus' or not next_node.layer.parameters:
            return False

        next_node.covered = True
        next_node.real_name = source_node.real_name
        B = next_node.layer.parameters[0].asarray()
        self.set_weight(source_node.name, 'bias', B)

        return True


    @staticmethod
    def _print_layer(source_node):
        print ("Layer: ", source_node.layer)
        print ("Parameters: ", source_node.layer.parameters)
        print ("Attributes: ", source_node.layer.attributes)
        for in_node in source_node.layer.inputs:
            print (in_node)


    def rename_UNKNOWN(self, source_node):
        print("Cntk Parser has not supported operator [%s] with name [%s]."
              % (source_node.type, source_node.name))

        self._print_layer(source_node)
        assert False


    @staticmethod
    def get_ndarray(variable):
        if variable.is_parameter:
            return variable.as_parameter().asarray()

        elif variable.is_constant:
            return variable.as_constant().asarray()

        else:
            raise ValueError("Unknown variable [{}].".format(variable))


    @staticmethod
    def _get_attribute(source_node, attribute_name):
        if attribute_name in source_node.attributes:
            return source_node.attributes

        node = source_node.block_root
        while not attribute_name in node.attributes:
            node = node.inputs[0].owner

        return node.attributes


    def rename_Convolution(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op="Conv")

        for input in source_node.layer.inputs:
            if input.name.endswith("W"):
                W = self.get_ndarray(input)
                break

        W = self.channel_first_conv_kernel_to_IR(W)
        self.set_weight(source_node.name, 'weights', W)

        attributes = CntkParser._get_attribute(source_node.layer, 'strides')

        kwargs = dict()
        kwargs['strides'] = [1] + list(attributes['strides'])[1:] + [1]
        kwargs['dilations'] = [1] + list(attributes['dilation'])[1:] + [1]
        kwargs['kernel_shape'] = list(W.shape)
        padding = attributes['autoPadding'][1:]

        for pad in padding:
            assert pad == padding[0]

        kwargs['auto_pad'] = 'SAME_LOWER' if padding[0] else 'VALID'
        kwargs['pads'] = self._convert_padding_to_IR(kwargs['kernel_shape'][:-2], padding)

        kwargs['use_bias'] = self._fuse_bias_node(source_node)

        assign_IRnode_values(IR_node, kwargs)


    def rename_ReLU(self, source_node):
        self._convert_identity_operation(source_node, new_op='Relu')


    def rename_Relu6(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_Plus(self, source_node):
        if not source_node.covered:
            assert not source_node.layer.parameters
            IR_node = self._convert_identity_operation(source_node, new_op='Add')


    def rename_Minus(self, source_node):
        if not source_node.covered:
            assert not source_node.layer.parameters
            self._convert_binary_operator(source_node, new_op='Sub')


    def rename_Sub(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_Reshape(self, source_node):
        IR_node = self._convert_identity_operation(source_node)
        new_shape = source_node.get_attr('newShape')
        kwargs = {'shape' : self.channel_first_shape_to_IR(new_shape)}
        assign_IRnode_values(IR_node, kwargs)


    def rename_Times(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='FullyConnected')

        W = source_node.layer.parameters[0].asarray().squeeze()
        self.set_weight(source_node.name, 'weights', W)

        kwargs = dict()
        kwargs['units'] = W.shape[-1]
        kwargs['use_bias'] = self._fuse_bias_node(source_node)
        assign_IRnode_values(IR_node, kwargs)


    def rename_MaxPooling(self, source_node):
        if source_node.layer.is_block:
            source_node.layer = source_node.layer.block_root.owner

        self.rename_Pooling(source_node)


    def rename_AveragePooling(self, source_node):
        self.rename_Pooling(source_node)

    def rename_Slice(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op = 'Slice')
        kwargs = dict()
        kwargs['axis'] = source_node.get_attr('axis')[-1] + 1
        kwargs['ends'] = source_node.get_attr('endIndex')
        kwargs['starts'] = source_node.get_attr('beginIndex')
        kwargs['strides'] = source_node.get_attr('sliceStrides')
        assign_IRnode_values(IR_node, kwargs)


    def rename_Splice(self, source_node):
        if len(source_node.in_edges) == 1:
            source_node.in_edges.append(source_node.in_edges[0])
        IR_node = self._convert_identity_operation(source_node, new_op='Concat')
        assign_IRnode_values(IR_node, {'axis' : source_node.get_attr('axis')[-1] + 1})


    def rename_StableSigmoid(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Sigmoid')

    def rename_BinaryCrossEntropy(self, source_node):
        pass

    def rename_Pooling(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='Pool')
        dim = len(IR_node.attr['_output_shapes'].list.shape[0].dim)
        kwargs = {}

        # strides
        kwargs['strides'] = list(source_node.get_attr('strides')) + [1]
        if len(kwargs['strides']) < dim:
            kwargs['strides'] = [1] + kwargs['strides']

        # window_shape
        kwargs['kernel_shape'] = list(source_node.get_attr('poolingWindowShape')) + [1]
        if len(kwargs['kernel_shape']) < dim:
            kwargs['kernel_shape'] = [1] + kwargs['kernel_shape']

        # pool type
        pool_type = source_node.get_attr('poolingType')
        if pool_type == _cntk.MAX_POOLING:
            kwargs['pooling_type'] = 'MAX'
        elif pool_type == _cntk.AVG_POOLING:
            kwargs['pooling_type'] = 'AVG'
        else:
            raise ValueError("Unknown pooling type [{}].".format(pool_type))

        # padding
        padding = source_node.get_attr('autoPadding')
        if len(padding) >= dim - 1:
            padding = padding[1:]
        elif len(padding) < dim - 2:
            padding.extend([padding[-1]] * (dim - len(padding) - 2))
        for pad in padding:
            assert pad == padding[-1]
        kwargs['auto_pad'] = 'SAME_LOWER' if padding[0] else 'VALID'
        kwargs['pads'] = self._convert_padding_to_IR(kwargs['kernel_shape'][1:-1], padding)

        assign_IRnode_values(IR_node, kwargs)


    def rename_DataInput(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='DataInput')
        shape = [-1] + list(source_node.layer.shape)
        assign_IRnode_values(IR_node, {'shape' : list_to_shape(self.channel_first_shape_to_IR(shape))})


    def rename_BatchNormalization(self, source_node):
        kwargs = dict()
        kwargs['scale'] = False
        kwargs['bias'] = False
        for param in source_node.layer.inputs:
            if param.name.endswith('scale'):
                self.set_weight(source_node.name, 'scale', self.get_ndarray(param).flatten())
                kwargs['scale'] = True

            elif param.name.endswith('bias'):
                self.set_weight(source_node.name, 'bias', self.get_ndarray(param).flatten())
                kwargs['bias'] = True

            elif param.name.lower().endswith('mean'):
                self.set_weight(source_node.name, 'mean', self.get_ndarray(param).flatten())

            elif param.name.lower().endswith('variance'):
                self.set_weight(source_node.name, 'var', self.get_ndarray(param).flatten())

        IR_node = self._convert_identity_operation(source_node, end_edge=1, new_op='BatchNorm')
        kwargs['epsilon'] = source_node.get_attr('epsilon')
        kwargs['axis'] = -1
        assign_IRnode_values(IR_node, kwargs)


    def _add_constant_node(self, constant_node, IR_node):
        new_node = self.IR_graph.node.add()
        new_node.name = constant_node.uid
        new_node.op = 'Constant'
        value = np.atleast_1d(self.get_ndarray(constant_node))
        self.set_weight(new_node.name, 'value', value)
        IR_node.input.append(new_node.name)


    def _convert_binary_operator(self, source_node, new_op):
        IR_node = self._convert_identity_operation(source_node, new_op=new_op)
        for in_node in source_node.layer.inputs:
            if in_node.is_constant:
                self._add_constant_node(in_node, IR_node)


    def rename_ElementTimes(self, source_node):
        if source_node.layer.inputs[0] == source_node.layer.inputs[1]:
            # TODO: Handle square
            pass

        self._convert_binary_operator(source_node, 'Mul')


    def rename_Log(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_Exp(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_Reciprocal(self, source_node):
        self._convert_identity_operation(source_node)


    def rename_Dropout(self, source_node):
        # self._print_layer(source_node)
        # print (source_node.name)
        # print (self.src_graph.get_parent(source_node.name, [0]).real_name)
        # assert False
        source_node.real_name = self.src_graph.get_parent(source_node.name, [0]).real_name


    def rename_Dense(self, source_node):
        IR_node = self._convert_identity_operation(source_node, new_op='FullyConnected', shape_transpose=False)
        for param in source_node.layer.inputs:
            if param.name.endswith('W'):
                w = np.squeeze(self.get_ndarray(param))
                if w.ndim > 2:
                    w = np.transpose(w, list(range(1, w.ndim - 1)) + [0, -1])
                    w = np.reshape(w, [-1, w.shape[-1]])
                self.set_weight(source_node.name, 'weights', w)
                assign_IRnode_values(IR_node, {'units' : w.shape[-1] })

            elif param.name.endswith('b'):
                self.set_weight(source_node.name, 'bias', self.get_ndarray(param))
                assign_IRnode_values(IR_node, {'use_bias' : True })


    def rename_Convolution2D(self, source_node):
        assert source_node.layer.is_block

        # Convolution
        kwargs = dict()
        conv_IR_node = self.IR_graph.node.add()
        conv_node = source_node.layer.block_root.inputs[0].owner.inputs[0].owner

        conv_IR_node.name = conv_node.uid
        conv_IR_node.op = 'Conv'
        conv_IR_node.input.append(self.get_parent(source_node.name, [0]).real_name)

        # Kernel
        conv_weight = source_node.layer.block_root.inputs[0].owner.inputs[0].owner.inputs[0]
        conv_weight = self.get_ndarray(conv_weight)
        W = self.channel_first_conv_kernel_to_IR(conv_weight)
        self.set_weight(conv_IR_node.name, 'weights', W)

        # Attributes
        conv_attr = source_node.layer.block_root.inputs[0].owner.inputs[0].owner.attributes

        kwargs['strides'] = [1] + list(conv_attr['strides'])[1:] + [1]
        kwargs['dilations'] = [1] + list(conv_attr['dilation'])[1:] + [1]
        kwargs['kernel_shape'] = list(W.shape)
        padding = conv_attr['autoPadding'][1:]

        for pad in padding:
            assert pad == padding[0]

        kwargs['auto_pad'] = 'SAME_LOWER' if padding[0] else 'VALID'
        kwargs['pads'] = self._convert_padding_to_IR(kwargs['kernel_shape'][:-2], padding)

        kwargs['use_bias'] = True

        assign_IRnode_values(conv_IR_node, kwargs)

        # Bias
        plus = source_node.layer.block_root.inputs[0].owner.inputs[1]
        plus = np.squeeze(self.get_ndarray(plus))
        self.set_weight(conv_IR_node.name, 'bias', plus)

        # Activation
        activation = source_node.layer.block_root.owner.op_name

        activation_IR = self.IR_graph.node.add()
        activation_IR.name = source_node.name
        activation_IR.input.append(conv_IR_node.name)
        if (activation == 'ReLU'):
            activation_IR.op = 'Relu'
        else:
            raise ValueError()


    def rename_Activation(self, source_node):
        assert source_node.layer.is_block

        op = source_node.layer.root_function.owner.name

        if op.startswith('relu'):
            new_op= 'Relu'
        else:
            raise ValueError()

        self._convert_identity_operation(source_node, new_op=new_op)


from __future__ import division
from __future__ import print_function

import numpy as np
import tensorflow
from tensorflow.python.framework import tensor_util
from tensorflow.core.framework import attr_value_pb2

import sys

from mmdnn.conversion.tensorflow.tensorflow_graph import *
import mmdnn.conversion.common.IR.graph_pb2 as graph_pb2
from mmdnn.conversion.rewriter.graph_matcher import *

from mmdnn.conversion.common.DataStructure import *
from tensorflow.core.framework.node_def_pb2 import NodeDef
from mmdnn.conversion.rewriter.rnn_utils import *


class UnitRewriterBase(object):


    def __init__(self, graph, weights_dict):
        self._graph = graph
        self._weights_dict = weights_dict

    def _rewrite_graph_by_pattern(self, pattern_name, graph_type):
        pattern = rnn_patterns[graph_type][pattern_name]
        matcher = GraphMatcher(pattern)
        match_results = list(matcher.match_ops(self._graph.get_nodes()))
        scope_names_dict = dict() # name: No.

        for i in range(len(match_results)):
            result = match_results[i]
            top_pattern_name = pattern_name + '_' + str(i)

            top_pattern = result._name_to_pattern[pattern_name]
            self.create_scope(result, top_pattern, scope_names_dict)

            top_op = result._pattern_to_op[top_pattern]
            top_op.scope = top_op.scope + '/top'

            # self.store_const_to_top(result)
            # self.set_top_node_prop(result, pattern_name)
            self.process_match_result(result, pattern_name)


    def rewrite_graph(self, pattern_names, graph_type):
        from six import string_types as _string_types
        if isinstance(pattern_names, _string_types):
            pattern_names = [pattern_names]
        elif not isinstance(pattern_names, list):
            raise ValueError
        for pattern_name in pattern_names:
            self._rewrite_graph_by_pattern(pattern_name, graph_type)

    def run(self, pattern_names, graph_type):
        self.rewrite_graph(pattern_names, graph_type)
    
    def store_const_to_top(self, match_result):
        top_node = list(match_result._pattern_to_op.values())[0]
        kwargs = dict()
        for pattern, op in match_result._pattern_to_op.items():
            if pattern.name and pattern.type == 'Const':
                if tensor_util.MakeNdarray(op.get_attr('value')).shape == (1, ):
                    kwargs[pattern.name] = np.asscalar(tensor_util.MakeNdarray(op.get_attr('value')))
                else:
                    kwargs[pattern.name] = np.squeeze(tensor_util.MakeNdarray(op.get_attr('value')))
        if hasattr(top_node, 'kwargs'):
            top_node.kwargs.update(kwargs)
        else:
            top_node.kwargs = kwargs

    def create_scope(self, result, pattern, scope_names_dict, parent_scope_name=''):
        op = result._pattern_to_op[pattern]

        if pattern.name:
            # Do not include input op.
            if 'input' in pattern.name.split('/')[-1]:
                return
            else:
                no = scope_names_dict.get(pattern.name, 0)
                scope_names_dict[pattern.name] = no + 1
                if parent_scope_name:
                    current_scope_name = '/'.join([parent_scope_name, pattern.name]) + '_' + str(no)
                else:
                    current_scope_name = pattern.name + '_' + str(no)
        else:
            current_scope_name = parent_scope_name
        op.scope = current_scope_name
        for sub_pattern in pattern.inputs:
            self.create_scope(result, sub_pattern, scope_names_dict, current_scope_name)

    def set_top_node_prop(self, match_result):
        raise NotImplementedError


    def process_match_result(self, match_result, pattern_name):
        raise NotImplementedError
